"""
2D visualization primitives based on Matplotlib.
1) Plot images with `plot_images`.
2) Call `plot_keypoints` or `plot_matches` any number of times.
3) Optionally: save a .png or .pdf plot (nice in papers!) with `save_plot`.
"""

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import numpy as np


def cm_RdGn(x):
    """Custom colormap: red (0) -> yellow (0.5) -> green (1)."""
    x = np.clip(x, 0, 1)[..., None]*2
    c = x*np.array([[0, 1., 0]]) + (2-x)*np.array([[1., 0, 0]])
    return np.clip(c, 0, 1)


def cm_BlRdGn(x_):
    """Custom colormap: blue (-1) -> red (0.0) -> green (1)."""
    x = np.clip(x_, 0, 1)[..., None]*2
    c = x*np.array([[0, 1., 0, 1.]]) + (2-x)*np.array([[1., 0, 0, 1.]])

    xn = -np.clip(x_, -1, 0)[..., None]*2
    cn = xn*np.array([[0, 0.1, 1, 1.]]) + (2-xn)*np.array([[1., 0, 0, 1.]])
    out = np.clip(np.where(x_[..., None] < 0, cn, c), 0, 1)
    return out


def cm_prune(x_):
    """ Custom colormap to visualize pruning """
    max_i = max(x_)
    norm_x = np.where(x_ == max_i, -1, (x_-1) / 9)
    return cm_BlRdGn(norm_x)


def plot_images(imgs, titles=None, cmaps='gray', dpi=100, pad=.5,
                adaptive=True):
    """Plot a set of images horizontally.
    Args:
        imgs: a list of NumPy or PyTorch images, RGB (H, W, 3) or mono (H, W).
        titles: a list of strings, as titles for each image.
        cmaps: colormaps for monochrome images.
        adaptive: whether the figure size should fit the image aspect ratios.
    """
    n = len(imgs)
    if not isinstance(cmaps, (list, tuple)):
        cmaps = [cmaps] * n

    if adaptive:
        ratios = [i.shape[1] / i.shape[0] for i in imgs]  # W / H
    else:
        ratios = [4/3] * n
    figsize = [sum(ratios)*4.5, 4.5]
    fig, ax = plt.subplots(
        1, n, figsize=figsize, dpi=dpi, gridspec_kw={'width_ratios': ratios})
    if n == 1:
        ax = [ax]
    for i in range(n):
        ax[i].imshow(imgs[i], cmap=plt.get_cmap(cmaps[i]))
        ax[i].get_yaxis().set_ticks([])
        ax[i].get_xaxis().set_ticks([])
        ax[i].set_axis_off()
        for spine in ax[i].spines.values():  # remove frame
            spine.set_visible(False)
        if titles:
            ax[i].set_title(titles[i])
    fig.tight_layout(pad=pad)


def plot_keypoints(kpts, colors='lime', ps=4, axes=None, a=1.0):
    """Plot keypoints for existing images.
    Args:
        kpts: list of ndarrays of size (N, 2).
        colors: string, or list of list of tuples (one for each keypoints).
        ps: size of the keypoints as float.
    """
    if not isinstance(colors, list):
        colors = [colors] * len(kpts)
    if not isinstance(a, list):
        a = [a] * len(kpts)
    if axes is None:
        axes = plt.gcf().axes
    for ax, k, c, alpha in zip(axes, kpts, colors, a):
        ax.scatter(k[:, 0], k[:, 1], c=c, s=ps, linewidths=0, alpha=alpha)


def plot_matches(kpts0, kpts1, color=None, lw=1.5, ps=4, a=1., labels=None,
                 axes=None):
    """Plot matches for a pair of existing images.
    Args:
        kpts0, kpts1: corresponding keypoints of size (N, 2).
        color: color of each match, string or RGB tuple. Random if not given.
        lw: width of the lines.
        ps: size of the end points (no endpoint if ps=0)
        indices: indices of the images to draw the matches on.
        a: alpha opacity of the match lines.
    """
    fig = plt.gcf()
    if axes is None:
        ax = fig.axes
        ax0, ax1 = ax[0], ax[1]
    else:
        ax0, ax1 = axes

    assert len(kpts0) == len(kpts1)
    if color is None:
        color = matplotlib.cm.hsv(np.random.rand(len(kpts0))).tolist()
    elif len(color) > 0 and not isinstance(color[0], (tuple, list)):
        color = [color] * len(kpts0)

    if lw > 0:
        for i in range(len(kpts0)):
            l = matplotlib.patches.ConnectionPatch(
                xyA=(kpts0[i, 0], kpts0[i, 1]), xyB=(kpts1[i, 0], kpts1[i, 1]),
                coordsA=ax0.transData, coordsB=ax1.transData,
                axesA=ax0, axesB=ax1,
                zorder=1, color=color[i], linewidth=lw, clip_on=True,
                alpha=a, label=None if labels is None else labels[i],
                picker=5.0)
            l.set_annotation_clip(True)
            fig.add_artist(l)

    # freeze the axes to prevent the transform to change
    ax0.autoscale(enable=False)
    ax1.autoscale(enable=False)

    if ps > 0:
        ax0.scatter(kpts0[:, 0], kpts0[:, 1], c=color, s=ps)
        ax1.scatter(kpts1[:, 0], kpts1[:, 1], c=color, s=ps)


def add_text(idx, text, pos=(0.01, 0.99), fs=15, color='w',
             lcolor='k', lwidth=2, ha='left', va='top'):
    ax = plt.gcf().axes[idx]
    t = ax.text(*pos, text, fontsize=fs, ha=ha, va=va,
                color=color, transform=ax.transAxes)
    if lcolor is not None:
        t.set_path_effects([
            path_effects.Stroke(linewidth=lwidth, foreground=lcolor),
            path_effects.Normal()])


def save_plot(path, **kw):
    """Save the current figure without any white margin."""
    plt.savefig(path, bbox_inches='tight', pad_inches=0, **kw)



from typing import Callable, Tuple
import torch
from torch import Tensor
import torch.nn.functional as F


def adj_to_list(
    x: Tensor, y: Tensor, mask: Tensor, padding: float = 0
) -> Tuple[Tensor, Tensor, Tensor]:
    x_pair = []
    y_pair = []
    for x1, x2, m in zip(x, y, mask):
        # print(x1.shape, x2.shape, m.shape)
        indices = torch.nonzero(m, as_tuple=True)
        i = indices[-2]
        j = indices[-1]
        if torch.numel(i) == 0 or torch.numel(j) == 0:
            x_pair.append(torch.full(x[0].size(), torch.nan, device=x.device))
            y_pair.append(torch.full(y[0].size(), torch.nan, device=y.device))
            continue
        x_pair.append(x1[..., i, :])
        y_pair.append(x2[..., j, :])
    x_pair = torch.nested.as_nested_tensor(x_pair, device=x.device).to_padded_tensor(  # type: ignore
        torch.nan
    )
    y_pair = torch.nested.as_nested_tensor(y_pair, device=y.device).to_padded_tensor(  # type: ignore
        torch.nan
    )
    mask = torch.isfinite(x_pair).all(dim=-1) & torch.isfinite(y_pair).all(dim=-1)  # type: ignore
    x_pair = x_pair.nan_to_num(padding)  # type: ignore
    y_pair = y_pair.nan_to_num(padding)  # type: ignore
    return x_pair, y_pair, mask  # type: ignore


def to_homogeneous(points: Tensor) -> Tensor:
    r"""Function that converts points from Euclidean to homogeneous space.

    Args:
        points: the points to be transformed with shape :math:`(*, N, D)`.

    Returns:
        the points in homogeneous coordinates :math:`(*, N, D+1)`.

    Examples:
        >>> input = tensor([[0., 0.]])
        >>> convert_points_to_homogeneous(input)
        tensor([[0., 0., 1.]])
    """
    if not isinstance(points, Tensor):
        raise TypeError(f"Input type is not a Tensor. Got {type(points)}")
    if len(points.shape) < 2:
        raise ValueError(f"Input must be at least a 2D tensor. Got {points.shape}")

    h_points = F.pad(points, [0, 1], "constant", 1.0)
    return h_points



import torch
from torch import Tensor
from .ransac import ransac
import ot  # type: ignore


def optimal_transport(xd: Tensor, yd: Tensor) -> Tensor:
    costs = torch.cdist(xd, yd, p=2)
    a = torch.ones(xd.shape[1], dtype=torch.float32, device=xd.device)
    a /= a.sum()

    b = torch.ones(yd.shape[1], dtype=torch.float32, device=xd.device)
    b /= b.sum()

    gs = []
    for i in range(xd.shape[0]):
        c = costs[i]
        g = ot.emd(a, b, c, numItermax=5000, numThreads="max")
        gs.append(g)
    gs = torch.stack(gs, dim=0)  # type: ignore
    gs = gs / gs.amax(dim=(-1, -2), keepdim=True)  # type: ignore
    return gs  # type: ignore


def match(xk: Tensor, xd: Tensor, yk: Tensor, yd: Tensor, **kwargs):
    w = optimal_transport(xd, yd)
    best_model, inliers, best_errors = ransac(xk, yk, w, **kwargs)
    return best_model, inliers, best_errors



import numpy as np
import faiss
import faiss.contrib.torch_utils
from prettytable import PrettyTable


def get_validation_recalls(r_list, q_list, k_values, gt, print_results=True, faiss_gpu=False, dataset_name='dataset without name ?'):
        
        embed_size = r_list.shape[1]
        if faiss_gpu:
            res = faiss.StandardGpuResources()
            flat_config = faiss.GpuIndexFlatConfig()
            flat_config.useFloat16 = True
            flat_config.device = 0
            faiss_index = faiss.GpuIndexFlatL2(res, embed_size, flat_config)
        # build index
        else:
            faiss_index = faiss.IndexFlatL2(embed_size)
        
        # add references
        faiss_index.add(r_list)

        # search for queries in the index
        _, predictions = faiss_index.search(q_list, max(k_values))
        
        
        
        # start calculating recall_at_k
        correct_at_k = np.zeros(len(k_values))
        for q_idx, pred in enumerate(predictions):
            for i, n in enumerate(k_values):
                # if in top N then also in top NN, where NN > N
                if np.any(np.in1d(pred[:n], gt[q_idx])):
                    correct_at_k[i:] += 1
                    break
        
        correct_at_k = correct_at_k / len(predictions)
        d = {k:v for (k,v) in zip(k_values, correct_at_k)}

        if print_results:
            print() # print a new line
            table = PrettyTable()
            table.field_names = ['K']+[str(k) for k in k_values]
            table.add_row(['Recall@K']+ [f'{100*v:.2f}' for v in correct_at_k])
            print(table.get_string(title=f"Performances on {dataset_name}"))
        
        return d



import torch

from ..utils.base_model import BaseModel


def find_nn(sim, ratio_thresh, distance_thresh):
    sim_nn, ind_nn = sim.topk(2 if ratio_thresh else 1, dim=-1, largest=True)
    dist_nn = 2 * (1 - sim_nn)
    mask = torch.ones(ind_nn.shape[:-1], dtype=torch.bool, device=sim.device)
    if ratio_thresh:
        mask = mask & (dist_nn[..., 0] <= (ratio_thresh**2)*dist_nn[..., 1])
    if distance_thresh:
        mask = mask & (dist_nn[..., 0] <= distance_thresh**2)
    matches = torch.where(mask, ind_nn[..., 0], ind_nn.new_tensor(-1))
    scores = torch.where(mask, (sim_nn[..., 0]+1)/2, sim_nn.new_tensor(0))
    return matches, scores


def mutual_check(m0, m1):
    inds0 = torch.arange(m0.shape[-1], device=m0.device)
    loop = torch.gather(m1, -1, torch.where(m0 > -1, m0, m0.new_tensor(0)))
    ok = (m0 > -1) & (inds0 == loop)
    m0_new = torch.where(ok, m0, m0.new_tensor(-1))
    return m0_new


class NearestNeighbor(BaseModel):
    default_conf = {
        'ratio_threshold': None,
        'distance_threshold': None,
        'do_mutual_check': True,
    }
    required_inputs = ['descriptors0', 'descriptors1']

    def _init(self, conf):
        pass

    def _forward(self, data):
        sim = torch.einsum(
            'bdn,bdm->bnm', data['descriptors0'], data['descriptors1'])
        matches0, scores0 = find_nn(
            sim, self.conf['ratio_threshold'], self.conf['distance_threshold'])
        matches1, scores1 = find_nn(
            sim.transpose(1, 2), self.conf['ratio_threshold'],
            self.conf['distance_threshold'])
        if self.conf['do_mutual_check']:
            matches1, scores1 = find_nn(
                sim.transpose(1, 2), self.conf['ratio_threshold'],
                self.conf['distance_threshold'])
            matches0 = mutual_check(matches0, matches1)
        return {
            'matches0': matches0,
            'matching_scores0': scores0,
        }



import argparse
import torch
from pathlib import Path
import h5py
import logging
from types import SimpleNamespace
import cv2
import numpy as np
from tqdm import tqdm
import pprint

import extractors
from unav.third_party.local_feature.SuperPoint_SuperGlue.base_model import dynamic_load
from tools import map_tensor


'''
A set of standard configurations that can be directly selected from the command
line using their name. Each is a dictionary with the following entries:
    - output: the name of the feature file that will be generated.
    - model: the model configuration, as passed to a feature extractor.
    - preprocessing: how to preprocess the images read from disk.
'''
confs = {
    'superpoint_aachen': {
        'output': 'feats-superpoint-n4096-r1024',
        'model': {
            'name': 'superpoint',
            'nms_radius': 3,
            'max_keypoints': 4096,
        },
        'preprocessing': {
            'grayscale': True,
            'resize_max': 1024,
        },
    },
    'superpoint_inloc': {
        'output': 'feats-superpoint-n4096-r1600',
        'model': {
            'name': 'superpoint',
            'nms_radius': 4,
            'max_keypoints': 4096,
        },
        'preprocessing': {
            'grayscale': True,
            'resize_max': 1600,
        },
    },
    'hfnet_superpoint': {
        'output': 'feats-superpoint',
        'model': {
            'name': 'superpoint',
            'nms_radius': 4,
            'max_keypoints': 4096,
        },
        'preprocessing': {
            'grayscale': True,
            'resize_max': 1600,
        },
    },
    'd2net-ss': {
        'output': 'feats-d2net-ss',
        'model': {
            'name': 'd2net',
            'multiscale': False,
        },
        'preprocessing': {
            'grayscale': False,
            'resize_max': 1600,
        },
    },
}


class ImageDataset(torch.utils.data.Dataset):
    default_conf = {
        'globs': ['*.jpg', '*.png', '*.jpeg', '*.JPG', '*.PNG'],
        'grayscale': False,
        'resize_max': None,
    }

    def __init__(self, root, conf):
        self.conf = conf = SimpleNamespace(**{**self.default_conf, **conf})
        self.root = root

        self.paths = []
        for g in conf.globs:
            self.paths += list(Path(root).glob('**/'+g))
        if len(self.paths) == 0:
            raise ValueError(f'Could not find any image in root: {root}.')
        self.paths = sorted(list(set(self.paths)))
        self.paths = [i.relative_to(root) for i in self.paths]
        logging.info(f'Found {len(self.paths)} images in root {root}.')

    def __getitem__(self, idx):
        path = self.paths[idx]
        if self.conf.grayscale:
            mode = cv2.IMREAD_GRAYSCALE
        else:
            mode = cv2.IMREAD_COLOR
        image = cv2.imread(str(self.root / path), mode)
        if not self.conf.grayscale:
            image = image[:, :, ::-1]  # BGR to RGB
        if image is None:
            raise ValueError(f'Cannot read image {str(path)}.')
        image = image.astype(np.float32)
        size = image.shape[:2][::-1]
        w, h = size

        if self.conf.resize_max and max(w, h) > self.conf.resize_max:
            scale = self.conf.resize_max / max(h, w)
            h_new, w_new = int(round(h*scale)), int(round(w*scale))
            image = cv2.resize(
                image, (w_new, h_new), interpolation=cv2.INTER_LINEAR)

        if self.conf.grayscale:
            image = image[None]
        else:
            image = image.transpose((2, 0, 1))  # HxWxC to CxHxW
        image = image / 255.

        data = {
            'name': path.as_posix(),
            'image': image,
            'original_size': np.array(size),
        }
        return data

    def __len__(self):
        return len(self.paths)

@torch.no_grad()
def main(conf, image_dir, export_dir, as_half=False):
    logging.info('Extracting local features with configuration:'
                 f'\n{pprint.pformat(conf)}')

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    Model = dynamic_load(extractors, conf['model']['name'])
    model = Model(conf['model']).eval().to(device)

    loader = ImageDataset(image_dir, conf['preprocessing'])
    loader = torch.utils.data.DataLoader(loader, num_workers=1)

    feature_path = Path(export_dir, conf['output']+'.h5')
    feature_path.parent.mkdir(exist_ok=True, parents=True)
    feature_file = h5py.File(str(feature_path), 'a')

    for data in tqdm(loader):
        pred = model(map_tensor(data, lambda x: x.to(device)))
        pred = {k: v[0].cpu().numpy() for k, v in pred.items()}

        pred['image_size'] = original_size = data['original_size'][0].numpy()
        if 'keypoints' in pred:
            size = np.array(data['image'].shape[-2:][::-1])
            scales = (original_size / size).astype(np.float32)
            pred['keypoints'] = (pred['keypoints'] + .5) * scales[None] - .5

        if as_half:
            for k in pred:
                dt = pred[k].dtype
                if (dt == np.float32) and (dt != np.float16):
                    pred[k] = pred[k].astype(np.float16)

        grp = feature_file.create_group(data['name'][0])
        for k, v in pred.items():
            grp.create_dataset(k, data=v)

        del pred

    feature_file.close()
    logging.info('Finished exporting features.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--image_dir', type=Path, required=True)
    parser.add_argument('--export_dir', type=Path, required=True)
    parser.add_argument('--conf', type=str, default='superpoint_aachen',
                        choices=list(confs.keys()))
    args = parser.parse_args()
    main(confs[args.conf], args.image_dir, args.export_dir)



import argparse
from os.path import dirname,join,exists,realpath
import yaml
import socket
from conversation import Server,ConnectedClient,utils
import loader
import torch

def get_server(root,map_data,hloc_config,server_config):
    return Server(root,map_data,hloc_config,server_config)

def main(root,hloc_config,server_config):
    map_data=loader.load_data(server_config)
    server = get_server(root,map_data,hloc_config,server_config)
    newConnectionsThread=server.set_new_connections(map_data)
    newConnectionsThread.start()

if __name__=='__main__':
    root = dirname(realpath(__file__)).replace('/src','')
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--server_config', type=str, default='configs/server.yaml')
    parser.add_argument('-l', '--hloc_config', type=str, default='configs/hloc.yaml')
    args = parser.parse_args()
    with open(args.hloc_config, 'r') as f:
        hloc_config = yaml.safe_load(f)
    with open(args.server_config, 'r') as f:
        server_config = yaml.safe_load(f)
    main(root,hloc_config,server_config)


import tkinter as tk
from tkinter import *
import numpy as np
from tkinter import ttk,filedialog
from tkinter.messagebox import showinfo
from ttkbootstrap import Style

import cv2
from PIL import Image, ImageTk, ImageDraw
import logging

from track import Hloc
from navigation import Trajectory,actions,command_debug, command_count
from visualization.destination_selection import Destination_window

import argparse
from os.path import dirname,join,exists,realpath
from os import listdir
import yaml
import loader

from tqdm import tqdm

class Main_window(ttk.Frame):
    image_types=['.jpg', '.png', '.jpeg', '.JPG', '.PNG','.HEIC']
    def __init__(self,master, map_data=None,hloc=None,trajectory=None,**config):
        ttk.Frame.__init__(self, master=master)
        self.config=config
        self.map_data=map_data
        self.map_scale=self.config['location']['scale']
        self.keyframe_name=map_data['keyframe_name']
        self.database_loc=map_data['database_loc']
        self.keyframe_location=map_data['keyframe_location']
        self.destination_list_name,self.destination_list_location=[],[]
        anchor_name=map_data['anchor_name']
        anchor_location=map_data['anchor_location']
        ind=0
        while anchor_name[ind][0]!='w':
            self.destination_list_name.append(anchor_name[ind])
            self.destination_list_location.append(anchor_location[ind])
            ind+=1
        self.hloc=hloc
        self.hloc.update_destination_map('New_York_City','LightHouse','3_small')
        self.trajectory=trajectory
        self.trajectory.update_destination_map('New_York_City','LightHouse','3_small')
        self.destination=[]
        self.initial=False
        self.halfway = False
        self.GT=None
        self.retrieval=True
        self.__layout_design()
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        console_format = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        console_handler.setFormatter(console_format)
        self.logger.addHandler(console_handler)

    def __layout_design(self):
        windowWidth = self.master.winfo_reqwidth()
        windowHeight = self.master.winfo_reqheight()
        self.positionRight = int(self.master.winfo_screenwidth() / 2 - windowWidth / 2)
        self.positionDown = int(self.master.winfo_screenheight() / 2 - windowHeight / 2)
        self.master.geometry("+{}+{}".format(self.positionRight, self.positionDown))
        self.master.title('Visualization')
        self.pack(side="left", fill="both", expand=False)
        self.master.geometry('2000x1200')
        self.master.columnconfigure(1, weight=1)
        self.master.columnconfigure(3, pad=7)
        self.master.rowconfigure(3, weight=1)
        self.master.rowconfigure(6, pad=7)
        """
        Localization image list
        """

        lbl = Label(self, text="Query frames:")
        lbl.grid(row=0, column=0, sticky=W, pady=4, ipadx=2)

        self.testing_image_folder=join(self.config['IO_root'],self.config['testing_image_folder'])
        ###
        self.testing_image_folder='/mnt/data/UNav-IO/logs/New_York_City/LightHouse/3_small/02526/images'
        ###
        print(self.testing_image_folder)

        self.query_names=sorted([file for file in listdir(self.testing_image_folder) if file.endswith('.png')])

        var2 = tk.StringVar()
        self.lb = tk.Listbox(self, listvariable=var2)

        self.scrollbar = Scrollbar(self, orient=VERTICAL)
        self.lb.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.lb.yview)

        self.lb.bind('<Double-Button-1>', lambda event, action='double':self.animate(action))
        self.lb.bind('<Up>', lambda event, action='up':
        self.animate(action))
        self.lb.bind('<Down>', lambda event, action='down':
        self.animate(action))
        self.testing_image_list = [i.split('.')[0] for i in sorted(self.query_names)]
        for i in self.testing_image_list:
            self.lb.insert('end', i)

        self.scrollbar.grid(row=7, column=1, columnspan=2, rowspan=4, padx=2, sticky='sn')

        """
        Function area
        """

        self.lb.grid(row=7, column=0, columnspan=1, rowspan=4, padx=2,
                     sticky=E + W + S + N)

        self.c=ttk.LabelFrame(self, text='Retrieval animation')
        self.c.grid(row=6, column=0, columnspan=2, padx=2,sticky=E + W)
        self.v1 = tk.IntVar()
        self.retrieval=True
        self.lb2 = tk.Radiobutton(self.c,
                                  text='Turn on',
                                  command=self.__retrieval_setting,
                                  variable=self.v1,
                                  value=0).pack(side=LEFT)
        self.lb2 = tk.Radiobutton(self.c,
                                  text='Turn off',
                                  command=self.__retrieval_setting,
                                  variable=self.v1,
                                  value=1).pack(side=RIGHT)
        self.shift = tk.Frame(self)
        self.shift.grid(row=14, column=0, padx=10, pady=1, sticky='we')
        scale = Label(self.shift, text='Retrieval number:')
        scale.pack(side=LEFT)
        self.e1 = tk.Entry(self.shift, width=3, justify='left')
        self.e1.pack(side=LEFT)
        self.e1.insert(END, '10')
        self.shift1 = tk.Frame(self)
        self.shift1.grid(row=12, column=0, padx=10, pady=1, sticky='we')
        scale = Label(self.shift1, text='Plotting scale:')
        scale.pack(side=LEFT)
        self.e2 = tk.Entry(self.shift1, width=5, justify='right')
        self.e2.insert(END, '0.1')
        self.e2.pack(side=LEFT)
        scale = Label(self.shift1, text='\'/pixel')
        scale.pack(side=LEFT)
        # ---------------------------------------------------
        separatorv1 = ttk.Separator(self, orient='vertical')
        separatorv1.grid(row=0, column=3, padx=10, columnspan=1, rowspan=70, sticky="sn")
        ebtn = tk.Button(self, text='Save Animation', width=16, command=self.gif_generator)
        ebtn.grid(row=15, column=0, padx=10, columnspan=1, rowspan=1)
        ebtn = tk.Button(self, text='Help', width=16, command=self.__help_info)
        ebtn.grid(row=16, column=0, padx=10, columnspan=1, rowspan=1)
        ebtn = tk.Button(self, text='Clear Destination', width=16, command=self.__clear_destination)
        ebtn.grid(row=17, column=0, padx=10, columnspan=1, rowspan=1)

        """
        Result area
        """

        location_config=self.config['location']
        floorplan_path=join(self.config['IO_root'],'data',location_config['place'],location_config['building'],location_config['floor'])

        fl_selected = filedialog.askopenfilename(initialdir=floorplan_path, title='Select Floor Plan')
        floorplan = Image.open(fl_selected)
        floorplan_draw = ImageDraw.Draw(floorplan)
        self.floorplan=floorplan.copy()
        
        for x,y in self.keyframe_location:
            floorplan_draw.ellipse((x - 2, y - 2, x + 2, y + 2), fill=(0, 255, 0))

        self.floorplan_with_keyframe = floorplan.copy()
        width, height = floorplan.size
        self.plot_scale =width/3400
        scale = 1600 / width
        newsize = (1600, int(height * scale))
        floorplan = floorplan.resize(newsize)
        tkimage1 = ImageTk.PhotoImage(floorplan)
        self.myvar1 = Label(self, image=tkimage1)
        self.myvar1.image = tkimage1
        self.myvar1.grid(row=0, column=4, columnspan=1, rowspan=40, sticky="snew")
        self.myvar1.bind('<Double-Button-1>', lambda event, action='double':
                            self.select_destination(action))
        
        self.length = 0

    def __clear_destination(self):
        self.destination=[]
        self.set_destination()

    def __retrieval_setting(self):
        if self.v1.get()==0:
            self.retrieval=True
        else:
            self.retrieval=False

    def prepare_image(self,image):
        draw=ImageDraw.Draw(image)
        l=60*self.plot_scale
        x_,y_=50*self.plot_scale,l
        ang=0
        x1, y1 = x_ - 40*self.plot_scale * np.sin(ang), y_ - 40*self.plot_scale * np.cos(ang)
        draw.ellipse((x_ - 20*self.plot_scale, y_ - 20*self.plot_scale, x_ + 20*self.plot_scale, y_ + 20*self.plot_scale), fill=(50, 0, 106))
        draw.line([(x_, y_), (x1, y1)], fill=(50, 0, 106), width=int(10*self.plot_scale))
        im=np.array(image)
        im=cv2.putText(im, 'Estimation pose', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                    1, (0, 0, 0), round(2*self.plot_scale), cv2.LINE_AA)
        image=Image.fromarray(im)
        draw = ImageDraw.Draw(image)
        if self.GT:
            l+=70*self.plot_scale
            x_, y_ = 50*self.plot_scale, l
            x1, y1 = x_ - 40*self.plot_scale * np.sin(ang), y_ - 40*self.plot_scale * np.cos(ang)
            draw.ellipse((x_ - 20*self.plot_scale, y_ - 20*self.plot_scale, x_ + 20*self.plot_scale, y_ + 20*self.plot_scale), fill=(255, 0, 255))
            draw.line([(x_, y_), (x1, y1)], fill=(255, 0, 255), width=int(10*self.plot_scale))
            im = np.array(image)
            im = cv2.putText(im, 'Ground truth pose', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                             1, (0, 0, 0), round(2**self.plot_scale), cv2.LINE_AA)
            image = Image.fromarray(im)
            draw = ImageDraw.Draw(image)
        if self.retrieval:
            l+=70*self.plot_scale
            x_, y_ = 50*self.plot_scale, l
            x1, y1 = x_ - 20*self.plot_scale * np.sin(ang), y_ - 20 *self.plot_scale* np.cos(ang)
            draw.ellipse((x_ - 10*self.plot_scale, y_ - 10*self.plot_scale, x_ + 10*self.plot_scale, y_ + 10*self.plot_scale), fill=(255, 0, 0))
            draw.line([(x_, y_), (x1, y1)], fill=(255, 0, 0), width=int(7*self.plot_scale))
            im = np.array(image)
            im = cv2.putText(im, 'Similar images', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                             1, (0, 0, 0), round(2*self.plot_scale), cv2.LINE_AA)
            image = Image.fromarray(im)
            draw = ImageDraw.Draw(image)
        if len(self.destination)>0:
            l+=70*self.plot_scale
            vertices = self.__star_vertices([50*self.plot_scale, l], 30)
            draw.polygon(vertices, fill='red')
            im = np.array(image)
            im = cv2.putText(im, 'Destination', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                             1, (0, 0, 0), round(2*self.plot_scale), cv2.LINE_AA)
            image = Image.fromarray(im)
            draw = ImageDraw.Draw(image)
        draw.rectangle([(10,5),(400+100*self.plot_scale,l+40*self.plot_scale)],outline='black',width=int(2*self.plot_scale))
        return draw,image
    
    def action(self,rot_ang,distance,floorplan):
        floorplan_numpy=np.array(floorplan)
        h,_,_=floorplan_numpy.shape
        rot_ang=(-rot_ang)%360
        rot_clock = round(rot_ang / 30) % 12
        floorplan_numpy = cv2.putText(floorplan_numpy, u'Please walk %.1f meters along %d clock\n' % (distance * self.scale, rot_clock), (10, h - 80), cv2.FONT_HERSHEY_SIMPLEX,
                         1.2, (255, 0, 0), 2, cv2.LINE_AA)
        print('Please walk %.1f meters along %d clock\n' % (distance * self.scale, rot_clock))
        return floorplan_numpy
    
    def gif(self):
        c,d=int(self.e4.get()),int(self.e5.get())
        dic_path = join(self.testing_image_folder, self.query_names[0])
        image = Image.open(dic_path)
        b, _ = image.size
        scale = c / b
        
        floorplan = self.floorplan.copy()
        draw_floorplan, floorplan = self.prepare_image(floorplan)
        e, _ = floorplan.size
        gif_shrink = e / d
        gif_ims=[]
        navigation_paths=[]

        for dic in tqdm(self.query_names,desc='Generate frames'):
            floorplan = self.floorplan.copy()
            draw_floorplan, floorplan = self.prepare_image(floorplan)
            dic_path = join(self.testing_image_folder, dic)
            image = Image.open(dic_path)
            query_image=image.copy()
            query_image_numpy=np.array(query_image)
            query_image_numpy=cv2.cvtColor(query_image_numpy,cv2.COLOR_BGR2RGB)
            pose = self.hloc.get_location(query_image_numpy)
            if pose:
                navigation_paths.append(pose[:2])
                if len(navigation_paths)>1:
                    x0,y0=navigation_paths[0]
                    for x1,y1 in navigation_paths[1:]:
                        draw_floorplan.line([(x1, y1), (x0, y0)], fill=(255, 0, 0), width=15)
                        x0,y0=x1,y1
                self.scale = float(self.e2.get())
                floorplan_numpy = np.array(floorplan)
                h, _, _ = floorplan_numpy.shape
                floorplan_numpy = cv2.putText(floorplan_numpy, 'Current location:  [%d,%d],  orientation:  %d degree' % (
                    pose[0], pose[1], pose[2]), (10, h - 140), cv2.FONT_HERSHEY_SIMPLEX,
                                1, (0, 0, 255), 2, cv2.LINE_AA)
                if len(self.destination)>0:
                    for d in self.destination:
                        xx,yy=self.keyframe_location[self.keyframe_name.index(d)]
                        floorplan_numpy = cv2.putText(floorplan_numpy, 'Destination location:  [%d,%d]' % (
                            xx, yy), (10, h - 200), cv2.FONT_HERSHEY_SIMPLEX,
                                        1, (0, 0, 0), 2, cv2.LINE_AA)
                floorplan = Image.fromarray(floorplan_numpy)
                draw_floorplan = ImageDraw.Draw(floorplan)
                # Check if self.hloc.pos exists and is not None
                if hasattr(self.hloc, 'indexes') and self.hloc.indexes is not None:
                    # print('orange')
                    # print(self.hloc.indexes)
                    # print(np.array(self.hloc.indexes).shape)
                    
                    # Iterate through each array of keypoints in self.hloc.pos
                    for i in self.hloc.indexes:
                        # Now iterate through each point in the current keypoints_array
                        x, y  = self.keyframe_location[i] 
                        # Adjust the size of the drawn ellipse based on your plot_scale
                        ellipse_bounds = [x - 3*self.plot_scale, y - 3*self.plot_scale, x + 3*self.plot_scale, y + 3*self.plot_scale]
                        # Draw the ellipse in orange to represent each keypoint
                        draw_floorplan.ellipse(ellipse_bounds, fill='orange')
                if len(self.destination)>0:
                    paths=[]
                    for d in self.destination:
                        xx,yy=self.keyframe_location[self.keyframe_name.index(d)]
                        vertices = self.__star_vertices([xx, yy], 30)
                        draw_floorplan.polygon(vertices, fill='red', outline='red')
                    for destination_id in self.destination:
                        path_list=self.trajectory.calculate_path(pose[:2], destination_id,self.hloc.get_current_floor())
                        if len(path_list)>0:
                            paths+=path_list

                    paths=[pose[:2]]+paths
                    for i in range(1,len(paths)):
                        x0, y0=paths[i-1]
                        x1, y1=paths[i]
                        vertices = self.__star_vertices([x0, y0], 15)
                        draw_floorplan.polygon(vertices, fill='yellow', outline='red')
                        draw_floorplan.line([(x0, y0), (x1, y1)], fill=(0,255,0), width=int(5*self.plot_scale))

                if self.retrieval:
                    for i in self.hloc.retrived_image_index:
                        x_,y_,ang=self.database_loc[i]
                        x1, y1 = x_ - 20*self.plot_scale * np.sin(ang), y_ - 20*self.plot_scale * np.cos(ang)
                        draw_floorplan.ellipse((x_ - 10*self.plot_scale, y_ - 10*self.plot_scale, x_ + 10*self.plot_scale, y_ + 10*self.plot_scale), fill=(255, 0, 0))
                        draw_floorplan.line([(x_, y_), (x1, y1)], fill=(255, 0, 0), width=int(7*self.plot_scale))
                draw_query = ImageDraw.Draw(query_image)
                for p in self.hloc.matched_2D:
                    xx, yy = p[0]-0.5, p[1]-0.5
                    draw_query.rectangle([(xx - 10, yy - 10), (xx + 10, yy + 10)], fill=(0, 0, 255))

                width_query, height_query = query_image.size
                newsize = (int(width_query * scale), int(height_query * scale))
                query_image = query_image.resize(newsize)
                draw_floorplan = ImageDraw.Draw(floorplan)
                matched_3D=np.array(self.hloc.matched_3D)
                self.matched_3D=np.ones((3,matched_3D.shape[0]))
                self.matched_3D[0,:]=matched_3D[:,0]
                self.matched_3D[1, :] = matched_3D[:, 2]
                matched_3D=(self.hloc.T@self.matched_3D).T
                for points in matched_3D:
                    x, y = points
                    draw_floorplan.rectangle([(x - 3, y - 3), (x + 3, y + 3)], fill=(0, 0, 255))

                width_query, height_query = query_image.size
                width_floorplan, height_floorplan = floorplan.size
                floorplan.paste(query_image, (width_floorplan - width_query, height_floorplan - height_query))

                for i, p in enumerate(self.hloc.matched_2D):
                    x2, y2 = (p[0]-0.5) * scale + width_floorplan - width_query, (p[1]-0.5) * scale + height_floorplan - height_query
                    x3, y3 = matched_3D[i]
                    draw_floorplan.line([(x3, y3), (x2, y2)], fill=(0, 255, 255), width=1)

                x1, y1 = pose[0] - 80 * np.sin(pose[2]/180*np.pi), pose[1] - 80 * np.cos(pose[2]/180*np.pi)
                draw_floorplan.ellipse((pose[0] - 40, pose[1] - 40, pose[0] + 40, pose[1] + 40), fill=(50, 0, 106))
                draw_floorplan.line([(pose[0], pose[1]), (x1, y1)], fill=(50, 0, 106), width=20)
                newsize = (int(width_floorplan / gif_shrink), int(height_floorplan / gif_shrink))
            else:
                floorplan = self.floorplan.copy()
                draw_floorplan, floorplan = self.prepare_image(floorplan)

                width_query, height_query = query_image.size
                newsize = (int(width_query * scale), int(height_query * scale))
                query_image = query_image.resize(newsize)

                width_query, height_query = query_image.size
                width_floorplan, height_floorplan = floorplan.size
                floorplan.paste(query_image, (width_floorplan - width_query, height_floorplan - height_query))
                width_floorplan, height_floorplan = floorplan.size
                newsize = (int(width_floorplan / gif_shrink), int(height_floorplan / gif_shrink))

            gif_ims.append(floorplan.resize(newsize))

        outf=str(self.e6.get())
        output_filename=join(self.config['IO_root'],self.config['testing_image_folder'], outf+'.mp4')
        out = cv2.VideoWriter(output_filename, 
                      cv2.VideoWriter_fourcc(*'mp4v'), 
                      float(self.e1.get()), 
                      newsize)
        
        
        for im in gif_ims:
            cv2_im = cv2.cvtColor(np.array(im), cv2.COLOR_RGB2BGR)
            out.write(cv2_im)

        out.release()

        self.win.destroy()
        showinfo("Saved!", "Animation saved to:\n{}".format(output_filename))

    def gif_generator(self):
        self.win = tk.Toplevel(self.master)
        self.win.wm_title("Warning!!!")
        self.win.geometry('330x130')
        self.win.geometry("+{}+{}".format(self.positionRight, self.positionDown))
        l = tk.Label(self.win, text="Save animation to outpath?")
        l.grid(row=1, column=0, columnspan=4, rowspan=2,
               padx=20, pady=30)

        b1 = ttk.Button(self.win, text="Yes", width=6, command=self.gif_config)
        b1.grid(row=3, column=0, columnspan=1, padx=40, rowspan=3)

        b2 = ttk.Button(self.win, text="No", width=6, command=self.win.destroy)
        b2.grid(row=3, column=1, columnspan=1, padx=40, rowspan=3)
    
    def gif_config(self):
        self.win.destroy()
        self.win = tk.Toplevel(self.master)
        self.win.wm_title("Animation setting")
        self.win.geometry('400x200')
        self.win.geometry("+{}+{}".format(self.positionRight, self.positionDown))
        self.shift = tk.Frame(self.win)
        self.shift.grid(row=0, column=0)
        self.shift1 = tk.Frame(self.shift)
        self.shift1.grid(row=0, column=0, pady=5, padx=60)
        scale = Label(self.shift1, text='Frame width in Gif:')
        scale.pack(side=LEFT)
        self.e4 = tk.Entry(self.shift1, width=4, justify='center')
        self.e4.insert(END, '1500')
        self.e4.pack(side=LEFT)
        self.shift2 = tk.Frame(self.shift)
        self.shift2.grid(row=1, column=0, pady=5, padx=60)
        scale = Label(self.shift2, text='Output gif width:')
        scale.pack(side=LEFT)
        self.e5 = tk.Entry(self.shift2, width=4, justify='center')
        self.e5.insert(END, '900')
        self.e5.pack(side=LEFT)
        self.shift4 = tk.Frame(self.shift)
        self.shift4.grid(row=2, column=0, pady=5, padx=60)
        scale = Label(self.shift4, text='fps:')
        scale.pack(side=LEFT)
        self.e1 = tk.Entry(self.shift4, width=2, justify='center')
        self.e1.insert(END, '1')
        self.e1.pack(side=LEFT)

        self.shift5 = tk.Frame(self.shift)
        self.shift5.grid(row=3, column=0, pady=5, padx=60)
        scale = Label(self.shift5, text='Gif name:')
        scale.pack(side=LEFT)
        self.e6 = tk.Entry(self.shift5, width=15, justify='center')
        # self.e6.insert(END, self.config['testing_image_folder'])
        self.e6.pack(side=LEFT)
        scale = Label(self.shift5, text='.mp4')
        scale.pack(side=LEFT)

        self.shift3 = tk.Frame(self.shift)
        self.shift3.grid(row=4, column=0, pady=5)
        b1 = ttk.Button(self.win, text="Go", width=6, command=self.gif)
        b1.grid(row=4, column=0, columnspan=1, padx=30, rowspan=1, sticky='w')
        b2 = ttk.Button(self.win, text="Cancel", width=6, command=self.win.destroy)
        b2.grid(row=4, column=0, columnspan=1, padx=20, rowspan=1, sticky='e')

    def __help_info(self):
        self.info = tk.Toplevel(self.master)
        self.info.geometry('800x650')
        self.info.title('Instruction')
        self.info.geometry("+{}+{}".format(self.positionRight - 300, self.positionDown - 200))
        # This will create a LabelFrame
        label_frame1 = LabelFrame(self.info, height=100, text='Steps')
        label_frame1.pack(expand='yes', fill='both')

        label1 = Label(label_frame1, text='1. Double click floor plan to pick a destination point.')
        label1.place(x=0, y=5)

        label2 = Label(label_frame1, text='2. Choose a query in query list to see localization.')
        label2.place(x=0, y=35)

        label3 = Label(label_frame1,
                       text='3. Click and see retrieved images and double click pairs to check local matches.')
        label3.place(x=0, y=65)

        label4 = Label(label_frame1, text='4. Repeat above to see other query results.')
        label4.place(x=0, y=95)

        label_frame1 = LabelFrame(self.info, height=400, text='Buttons')
        label_frame1.pack(expand='yes', fill='both', side='bottom')

        label_1 = LabelFrame(label_frame1, height=60, text='Retrieval animation')
        label_1.place(x=5, y=23)
        label1 = Label(label_1, text='Turn on or turn off the display of retrieval images')
        label1.pack()

        label_2 = LabelFrame(label_frame1, height=40, text='Plotting scale')
        label_2.place(x=5, y=70)
        label2 = Label(label_2, text='How many foot of per pixel.')
        label2.pack()

        label_3 = LabelFrame(label_frame1, height=40, text='Retrieval number')
        label_3.place(x=5, y=117)
        label3 = Label(label_3,
                       text='How many retrieved images to use.')
        label3.pack()

        label_4 = LabelFrame(label_frame1, height=60, text='Save Animation')
        label_4.place(x=5, y=164)
        label4 = Label(label_4,
                       text='Save the localization results.')
        label4.pack()
    
    def __star_vertices(self,center,r):
        out_vertex = [(r*self.plot_scale * np.cos(2 * np.pi * k / 5 + np.pi / 2- np.pi / 5) + center[0],
                       r*self.plot_scale * np.sin(2 * np.pi * k / 5 + np.pi / 2- np.pi / 5) + center[1]) for k in range(5)]
        r = r/2
        in_vertex = [(r*self.plot_scale * np.cos(2 * np.pi * k / 5 + np.pi / 2 ) + center[0],
                      r*self.plot_scale * np.sin(2 * np.pi * k / 5 + np.pi / 2 ) + center[1]) for k in range(5)]
        vertices = []
        for i in range(5):
            vertices.append(out_vertex[i])
            vertices.append(in_vertex[i])
        vertices = tuple(vertices)
        return vertices
    
    def __visualize_result(self,floorplan_with_keyframe):
        draw_floorplan_with_keyframe=ImageDraw.Draw(floorplan_with_keyframe)
        # Check if self.hloc.pos exists and is not None
        if hasattr(self.hloc, 'indexes') and self.hloc.indexes is not None:
            # print('orange')
            # print(self.hloc.indexes)
            # print(np.array(self.hloc.indexes).shape)
            
            # Iterate through each array of keypoints in self.hloc.pos
            for i in self.hloc.indexes:
                # Now iterate through each point in the current keypoints_array
                x, y  = self.keyframe_location[i] 
                # Adjust the size of the drawn ellipse based on your plot_scale
                ellipse_bounds = [x - 3*self.plot_scale, y - 3*self.plot_scale, x + 3*self.plot_scale, y + 3*self.plot_scale]
                # Draw the ellipse in orange to represent each keypoint
                draw_floorplan_with_keyframe.ellipse(ellipse_bounds, fill='orange')
        floorplan_with_keyframe.save("floorplan_with_keypoints.png")

        l=60*self.plot_scale
        x_,y_=50*self.plot_scale,l
        ang=0
        x1, y1 = x_ - 40*self.plot_scale * np.sin(ang), y_ - 40*self.plot_scale * np.cos(ang)
        draw_floorplan_with_keyframe.ellipse((x_ - 20*self.plot_scale, y_ - 20*self.plot_scale, x_ + 20*self.plot_scale, y_ + 20*self.plot_scale), fill=(50, 0, 106))
        draw_floorplan_with_keyframe.line([(x_, y_), (x1, y1)], fill=(50, 0, 106), width=int(10*self.plot_scale))
        floorplan_with_keyframe_np=np.array(floorplan_with_keyframe)
        h, _, _ = floorplan_with_keyframe_np.shape 
        floorplan_with_keyframe_np=cv2.putText(floorplan_with_keyframe_np, 'Estimation pose', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                    1, (0, 0, 0), round(2*self.plot_scale), cv2.LINE_AA)
        floorplan_with_keyframe=Image.fromarray(floorplan_with_keyframe_np)
        draw_floorplan_with_keyframe = ImageDraw.Draw(floorplan_with_keyframe)
        if self.GT:
            l+=70*self.plot_scale
            x_, y_ = 50*self.plot_scale, l
            x1, y1 = x_ - 40*self.plot_scale * np.sin(ang), y_ - 40*self.plot_scale * np.cos(ang)
            draw_floorplan_with_keyframe.ellipse((x_ - 20*self.plot_scale, y_ - 20*self.plot_scale, x_ + 20*self.plot_scale, y_ + 20*self.plot_scale), fill=(255, 0, 255))
            draw_floorplan_with_keyframe.line([(x_, y_), (x1, y1)], fill=(255, 0, 255), width=int(10*self.plot_scale))
            floorplan_with_keyframe_np = np.array(floorplan_with_keyframe)
            floorplan_with_keyframe_np = cv2.putText(floorplan_with_keyframe_np, 'Ground truth pose', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                             1, (0, 0, 0), round(2**self.plot_scale), cv2.LINE_AA)
            floorplan_with_keyframe = Image.fromarray(floorplan_with_keyframe_np)
            draw_floorplan_with_keyframe = ImageDraw.Draw(floorplan_with_keyframe)
            ang_gt = self.GT[dic.split('.')[0]]['rot']
            x_gt, y_gt = self.GT[dic.split('.')[0]]['trans'][0], self.GT[dic.split('.')[0]]['trans'][1]
            x1, y1 = x_gt - 40 * np.sin(ang_gt), y_gt - 40 * np.cos(ang_gt)
            draw_floorplan_with_keyframe.ellipse((x_gt - 20, y_gt - 20, x_gt + 20, y_gt + 20), fill=(255, 0, 255))
            draw_floorplan_with_keyframe.line([(x_gt, y_gt), (x1, y1)], fill=(255, 0, 255), width=10)

        if self.retrieval:
            l+=70*self.plot_scale
            x_, y_ = 50*self.plot_scale, l
            x1, y1 = x_ - 20*self.plot_scale * np.sin(ang), y_ - 20 *self.plot_scale* np.cos(ang)
            draw_floorplan_with_keyframe.ellipse((x_ - 10*self.plot_scale, y_ - 10*self.plot_scale, x_ + 10*self.plot_scale, y_ + 10*self.plot_scale), fill=(255, 0, 0))
            draw_floorplan_with_keyframe.line([(x_, y_), (x1, y1)], fill=(255, 0, 0), width=int(7*self.plot_scale))
            floorplan_with_keyframe_np = np.array(floorplan_with_keyframe)
            floorplan_with_keyframe_np = cv2.putText(floorplan_with_keyframe_np, 'Similar images', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                             1, (0, 0, 0), round(2*self.plot_scale), cv2.LINE_AA)
            floorplan_with_keyframe = Image.fromarray(floorplan_with_keyframe_np)
            draw_floorplan_with_keyframe = ImageDraw.Draw(floorplan_with_keyframe)
        if len(self.destination)>0:
            l+=70*self.plot_scale
            vertices = self.__star_vertices([50*self.plot_scale, l], 30)
            draw_floorplan_with_keyframe.polygon(vertices, fill='red')
            floorplan_with_keyframe_np = np.array(floorplan_with_keyframe)
            floorplan_with_keyframe_np = cv2.putText(floorplan_with_keyframe_np, 'Destination', (int(100*self.plot_scale), int(l)), cv2.FONT_HERSHEY_SIMPLEX,
                             1, (0, 0, 0), round(2*self.plot_scale), cv2.LINE_AA)
            
            floorplan_with_keyframe = Image.fromarray(floorplan_with_keyframe_np)
            draw_floorplan_with_keyframe = ImageDraw.Draw(floorplan_with_keyframe)
        draw_floorplan_with_keyframe.rectangle([(10,5),(400+100*self.plot_scale,l+40*self.plot_scale)],outline='black',width=int(2*self.plot_scale))     

        if self.pose:
            x,y,ang=self.pose
            ang_pi=ang/180*np.pi
            x1, y1 = x - 40*self.plot_scale * np.sin(ang_pi), y - 40*self.plot_scale * np.cos(ang_pi)
            draw_floorplan_with_keyframe.ellipse((x - 20*self.plot_scale, y - 20*self.plot_scale, x + 20*self.plot_scale, y + 20*self.plot_scale), fill=(50, 0, 106))
            draw_floorplan_with_keyframe.line([(x, y), (x1, y1)], fill=(50, 0, 106), width=int(10*self.plot_scale))
            message='Current location:  [%d,%d],  orientation:  %d degree' % (
                x, y, ang)
        else:
            message= 'Cannot localize'
        
        floorplan_with_keyframe_np = np.array(floorplan_with_keyframe) 
            
        floorplan_with_keyframe_np = cv2.putText(floorplan_with_keyframe_np, message, (10, h - 200), cv2.FONT_HERSHEY_SIMPLEX,
                            1, (0, 0, 255), 2, cv2.LINE_AA)
        if self.pose:
            floorplan_with_keyframe_np = cv2.putText(floorplan_with_keyframe_np, self.instruction_message, (10, h - 80), cv2.FONT_HERSHEY_SIMPLEX,
                        1.2, (255, 0, 0), 2, cv2.LINE_AA)
                        

        if len(self.destination)>0:
            for ind,d in enumerate(self.destination):
                x_,y_=self.destination_list_location[self.destination_list_name.index(d)]
                floorplan_with_keyframe_np=cv2.putText(floorplan_with_keyframe_np, 'Destination location %d:  [%d,%d]' % (ind+1,
                            x_,y_), (10, h - 140-(len(self.destination)-ind-1)*60), cv2.FONT_HERSHEY_SIMPLEX,
                                            1, (0, 0, 0), 2, cv2.LINE_AA)

        floorplan_with_keyframe = Image.fromarray(floorplan_with_keyframe_np)
        draw_floorplan_with_keyframe = ImageDraw.Draw(floorplan_with_keyframe)

        if self.pose and len(self.destination)>0:
            paths=[self.pose[:2]]+self.paths
            for i in range(1,len(paths)):
                x0, y0=paths[i-1]
                x1, y1=paths[i]
                draw_floorplan_with_keyframe.line([(x0, y0), (x1, y1)], fill=(255,0,0), width=int(10*self.plot_scale))
                distance=np.linalg.norm([x1-x0,y1-y0])
                rot=np.arctan2(x1-x0,y1-y0)
                rot_ang=(rot-ang)/np.pi*180

        if self.retrieval:
            for i in self.hloc.retrived_image_index:
                x_,y_,ang=self.database_loc[i]
                x1, y1 = x_ - 20*self.plot_scale * np.sin(ang), y_ - 20*self.plot_scale * np.cos(ang)
                draw_floorplan_with_keyframe.ellipse((x_ - 10*self.plot_scale, y_ - 10*self.plot_scale, x_ + 10*self.plot_scale, y_ + 10*self.plot_scale), fill=(255, 0, 0))
                draw_floorplan_with_keyframe.line([(x_, y_), (x1, y1)], fill=(255, 0, 0), width=int(7*self.plot_scale))
        del self.hloc.retrived_image_index
        
        width, height = floorplan_with_keyframe.size
        scale = 1600 / width
        newsize = (1600, int(height * scale))
        floorplan_with_keyframe = floorplan_with_keyframe.resize(newsize)
        tkimage1 = ImageTk.PhotoImage(floorplan_with_keyframe)
        self.myvar1 = Label(self, image=tkimage1)
        self.myvar1.image = tkimage1
        self.myvar1.grid(row=0, column=4, columnspan=1, rowspan=40, sticky="snew")
        self.myvar1.bind('<Double-Button-1>', lambda event, action='double':
        self.select_destination(action))

    def set_destination(self):
        floorplan=self.floorplan.copy()
        draw_floorplan = ImageDraw.Draw(floorplan)
        for i,keyframe_name in enumerate(self.keyframe_name):
            if keyframe_name not in self.destination:
                x,y=self.keyframe_location[i]
                draw_floorplan.ellipse((x - 2*self.plot_scale, y - 2*self.plot_scale, x + 2*self.plot_scale, y + 2*self.plot_scale), fill=(0, 255, 0))

        for d in self.destination:
            x, y =self.keyframe_location[self.keyframe_name.index(d)]
            vertices = self.__star_vertices([x, y],30)
            draw_floorplan.polygon(vertices, fill='red')

        self.floorplan_with_keyframe = floorplan.copy()

        width, height = floorplan.size
        scale = 1600 / width
        newsize = (1600, int(height * scale))
        floorplan = floorplan.resize(newsize)
        tkimage1 = ImageTk.PhotoImage(floorplan)
        self.myvar1 = Label(self, image=tkimage1)
        self.myvar1.image = tkimage1
        self.myvar1.grid(row=0, column=4, columnspan=1, rowspan=40, sticky="snew")
        self.myvar1.bind('<Double-Button-1>', lambda event, action='double':
        self.select_destination(action))

    def animate(self,action):
        """
        Show testing image in GUI
        """
        self.value = self.lb.get(self.lb.curselection())
        if action == 'up':
            i = self.testing_image_list.index(self.value)
            if i > 0:
                self.value = self.testing_image_list[i - 1]
        if action == 'down':
            i = self.testing_image_list.index(self.value)
            if i < (len(self.testing_image_list) - 1):
                self.value = self.testing_image_list[i + 1]
        for type in self.image_types:
            q_path=join(self.testing_image_folder, self.value + type)
            if exists(q_path):
                testing_image = Image.open(q_path)
                break
        
        width, height = testing_image.size
        scale = 210 / width
        newsize = (210, int(height * scale))

        timg=testing_image.copy()
        timg = timg.resize(newsize)
        tkimage1 = ImageTk.PhotoImage(timg)
        self.myvar1 = Label(self, image=tkimage1)
        self.myvar1.image = tkimage1
        self.myvar1.grid(row=1, column=0, columnspan=1, padx=10, rowspan=5, sticky="snew")

        """
        Localize testing image
        """
        scale = 640 / width
        newsize = (640, int(height * scale))
        testing_image = testing_image.resize(newsize)
        testing_image=np.array(testing_image)
        testing_image=cv2.cvtColor(testing_image,cv2.COLOR_BGR2RGB)

        self.pose = self.hloc.get_location(testing_image) # Localize image

        """
        Navigation
        """
        ###
        self.destination=['14843']
        print("self.pose:")
        print(self.pose)
        ###
        if self.pose and len(self.destination)>0:
            self.paths=[]
            for destination_id in self.destination:
                path_list=self.trajectory.calculate_path(self.pose[:2], destination_id,self.hloc.get_current_floor())
                if len(path_list)>0:
                    self.paths+=path_list
            print('#########################')
            print(self.paths)
            action_list=actions(self.pose,self.paths,self.map_scale)
            print(action_list)
            
            if len(action_list)==0:
                self.instruction_message="There is no path to the destination. "
            else:
                length = action_list[0][1]
                if not self.initial:
                    self.instruction_message=command_debug(action_list)
                    self.initial = True
                    self.base_len=length
                else:
                    # try:
                    self.instruction_message=command_count(self,action_list,length)
                    # except:
                    print(action_list)

                
            x, y, an = self.pose
            print("x="+str(x)+", y="+str(y)+", angle="+str(an))
            self.logger.info(f"===============================================\n                                                       {self.instruction_message}\n                                                       ===============================================")
        elif not self.pose:
            self.instruction_message = 'Cannot localize at this point.'
            self.logger.info(f"===============================================\n                                                       {self.instruction_message}\n                                                       ===============================================")


        """
        Animate results on the floor plan
        """
        floorplan_with_keyframe=self.floorplan_with_keyframe.copy()
        self.__visualize_result(floorplan_with_keyframe)

    def select_destination(self,w):
        self.newWindow = tk.Toplevel(self.master)
        self.app1 = Destination_window(self.newWindow, parent=self)

def main(root,hloc_config,visual_config):
    map_data=loader.load_data(visual_config)
    hloc = Hloc(root, map_data, hloc_config)
    trajectory=Trajectory(map_data)

    style = Style(theme='darkly')
    root = style.master
    Main_window(root, map_data=map_data,hloc=hloc,trajectory=trajectory,**visual_config)
    root.mainloop()

if __name__=='__main__':
    root = dirname(realpath(__file__)).replace('/src','')
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--visual_config', type=str, default='configs/visualization.yaml')
    parser.add_argument('-l', '--hloc_config', type=str, default='configs/hloc.yaml')
    args = parser.parse_args()
    with open(args.hloc_config, 'r') as f:
        hloc_config = yaml.safe_load(f)
    with open(args.visual_config, 'r') as f:
        visual_config = yaml.safe_load(f)
    input('press enter to continue ...')
    main(root,hloc_config,visual_config)


from pathlib import Path
import torch
import cv2
import numpy as np
from typing import Union, List, Optional


def read_image(path: Path, grayscale: bool = False) -> np.ndarray:
    """Read an image from path as RGB or grayscale"""
    mode = cv2.IMREAD_GRAYSCALE if grayscale else cv2.IMREAD_COLOR
    image = cv2.imread(str(path), mode)
    if image is None:
        raise IOError(f'Could not read image at {path}.')
    if not grayscale:
        image = image[..., ::-1]
    return image


def numpy_image_to_torch(image: np.ndarray) -> torch.Tensor:
    """Normalize the image tensor and reorder the dimensions."""
    if image.ndim == 3:
        image = image.transpose((2, 0, 1))  # HxWxC to CxHxW
    elif image.ndim == 2:
        image = image[None]  # add channel axis
    else:
        raise ValueError(f'Not an image: {image.shape}')
    return torch.tensor(image / 255., dtype=torch.float)


def resize_image(image: np.ndarray, size: Union[List[int], int],
                 fn: str, interp: Optional[str] = 'area') -> np.ndarray:
    """Resize an image to a fixed size, or according to max or min edge."""
    h, w = image.shape[:2]

    fn = {'max': max, 'min': min}[fn]
    if isinstance(size, int):
        scale = size / fn(h, w)
        h_new, w_new = int(round(h*scale)), int(round(w*scale))
        scale = (w_new / w, h_new / h)
    elif isinstance(size, (tuple, list)):
        h_new, w_new = size
        scale = (w_new / w, h_new / h)
    else:
        raise ValueError(f'Incorrect new size: {size}')
    mode = {
        'linear': cv2.INTER_LINEAR,
        'cubic': cv2.INTER_CUBIC,
        'nearest': cv2.INTER_NEAREST,
        'area': cv2.INTER_AREA}[interp]
    return cv2.resize(image, (w_new, h_new), interpolation=mode), scale


def load_image(path: Path, grayscale: bool = False, resize: int = None,
               fn: str = 'max', interp: str = 'area') -> torch.Tensor:
    img = read_image(path, grayscale=grayscale)
    scales = [1, 1]
    if resize is not None:
        img, scales = resize_image(img, resize, fn=fn, interp=interp)
    return numpy_image_to_torch(img), torch.Tensor(scales)


def match_pair(extractor, matcher, image0, image1, scales0=None, scales1=None):
    device = image0.device
    data = {'image0': image0[None].cuda(), 'image1': image1[None].cuda()}
    img0, img1 = data['image0'], data['image1']

    feats0, feats1 = extractor({'image': img0}), extractor({'image': img1})
    pred = {**{k+'0': v for k, v in feats0.items()},
            **{k+'1': v for k, v in feats1.items()},
            **data}

    pred = {**pred, **matcher(pred)}
    pred = {k: v.to(device).detach()[0] if
            isinstance(v, torch.Tensor) else v for k, v in pred.items()}
    if scales0 is not None:
        pred['keypoints0'] = (pred['keypoints0'] + 0.5) / scales0[None] - 0.5
    if scales1 is not None:
        pred['keypoints1'] = (pred['keypoints1'] + 0.5) / scales1[None] - 0.5
    del feats0, feats1
    torch.cuda.empty_cache()

    # create match indices
    kpts0, kpts1 = pred['keypoints0'], pred['keypoints1']
    matches0, mscores0 = pred['matches0'], pred['matching_scores0']
    valid = matches0 > -1
    matches = torch.stack([torch.where(valid)[0], matches0[valid]], -1)
    return {**pred, 'matches': matches, 'matching_scores': mscores0[valid]}



from pathlib import Path
from types import SimpleNamespace
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
from einops import rearrange, repeat
from typing import Optional, List, Callable

try:
    from flash_attn.modules.mha import FlashCrossAttention
    FOUND_OFFICIAL_FLASH = True
except ModuleNotFoundError:
    FOUND_OFFICIAL_FLASH = False

torch.backends.cudnn.deterministic = True


@torch.cuda.amp.custom_fwd(cast_inputs=torch.float32)
def normalize_keypoints(
        kpts: torch.Tensor,
        size: Optional[List[int]] = None,
        shape: Optional[List[int]] = None) -> torch.Tensor:
    if size is None:
        assert shape is not None
        _, _, h, w = shape
        one = kpts.new_tensor(1)
        size = torch.stack([one*w, one*h])[None]
    shift = size.float().to(kpts) / 2
    scale = size.max(1).values.float().to(kpts) / 2
    kpts = (kpts - shift[:, None]) / scale[:, None, None]
    return kpts


def rotate_half(x: torch.Tensor) -> torch.Tensor:
    x = rearrange(x, '... (d r) -> ... d r', r=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, '... d r -> ... (d r)')

# def apply_cached_rotary_emb(
#         freqs: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
#     return (t * freqs[0]) + (rotate_half(t) * freqs[1])

def apply_cached_rotary_emb(
        freqs: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
    t_rot=rotate_half(t).permute(1,0,2,3)
    t=t.permute(1,0,2,3)
    return ((t * freqs[0]) + (t_rot * freqs[1])).permute(1,0,2,3)


class LearnableFourierPositionalEncoding(nn.Module):
    def __init__(self, M: int, dim: int, F_dim: int = None, gamma: float = 1.0) -> None:
        super().__init__()
        F_dim = F_dim if F_dim is not None else dim
        self.gamma = gamma
        self.Wr = nn.Linear(M, F_dim // 2, bias=False)
        nn.init.normal_(self.Wr.weight.data, mean=0, std=self.gamma ** -2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ encode position vector """
        projected = self.Wr(x)
        cosines, sines = torch.cos(projected), torch.sin(projected)
        emb = torch.stack([cosines, sines], 0).unsqueeze(-4)
        return repeat(emb, '... n -> ... (n r)', r=2)


class TokenConfidence(nn.Module):
    def __init__(self, dim: int) -> None:
        super(TokenConfidence, self).__init__()
        self.token = nn.Sequential(
            nn.Linear(dim, 1),
            nn.Sigmoid()
        )

    def forward(self, desc0: torch.Tensor, desc1: torch.Tensor):
        """ get confidence tokens """
        return (
            self.token(desc0.detach().float()).squeeze(-1),
            self.token(desc1.detach().float()).squeeze(-1))


class FastAttention(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.s = dim ** -0.5

    def forward(self, q, k, v) -> torch.Tensor:
        if hasattr(F, 'scaled_dot_product_attention'):
            q, k, v = [x.contiguous() for x in [q, k, v]]
            return F.scaled_dot_product_attention(q, k, v)
        else:
            s = self.s
            attn = F.softmax(torch.einsum('...id,...jd->...ij', q, k) * s, -1)
            return torch.einsum('...ij,...jd->...id', attn, v)


class FlashAttention(nn.Module):
    def __init__(self, *args) -> None:
        super().__init__()
        if FOUND_OFFICIAL_FLASH:
            self.flash = FlashCrossAttention()

    def forward(self, q, k, v) -> torch.Tensor:
        if FOUND_OFFICIAL_FLASH:
            q, k, v = [x.transpose(-2, -3) for x in [q, k, v]]
            m = self.flash(q.half(), torch.stack([k, v], 2).half())
            return m.transpose(-2, -3).to(q.dtype)
        else:
            args = [x.half().contiguous() for x in [q, k, v]]
            return F.scaled_dot_product_attention(*args).to(q.dtype)


class Transformer(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, flash: bool = False, bias: bool = True) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        assert self.embed_dim % num_heads == 0
        self.head_dim = self.embed_dim // num_heads
        self.Wqkv = nn.Linear(embed_dim, 3*embed_dim, bias=bias)
        attn = FlashAttention if flash else FastAttention
        self.inner_attn = attn(self.head_dim)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.ffn = nn.Sequential(
            nn.Linear(2*embed_dim, 2*embed_dim),
            nn.LayerNorm(2*embed_dim, elementwise_affine=True),
            nn.GELU(),
            nn.Linear(2*embed_dim, embed_dim)
        )

    def _forward(self, x: torch.Tensor, encoding: Optional[torch.Tensor] = None):
        qkv = self.Wqkv(x)
        qkv = rearrange(qkv, 'b n (h d three) -> b h n d three',
                        three=3, h=self.num_heads)
        q, k, v = qkv[..., 0], qkv[..., 1], qkv[..., 2]
        if encoding is not None:
            q = apply_cached_rotary_emb(encoding, q)
            k = apply_cached_rotary_emb(encoding, k)
        context = self.inner_attn(q, k, v)
        message = self.out_proj(rearrange(context, 'b h n d -> b n (h d)'))
        return x + self.ffn(torch.cat([x, message], -1))

    def forward(self, x0, x1, encoding0=None, encoding1=None):
        return self._forward(x0, encoding0), self._forward(x1, encoding1)


class CrossTransformer(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, flash: bool = False, bias: bool = True) -> None:
        super().__init__()
        self.heads = num_heads
        dim_head = embed_dim // num_heads
        self.scale = dim_head ** -0.5
        inner_dim = dim_head * num_heads
        self.to_qk = nn.Linear(embed_dim, inner_dim, bias=bias)
        self.to_v = nn.Linear(embed_dim, inner_dim, bias=bias)
        self.to_out = nn.Linear(inner_dim, embed_dim, bias=bias)
        self.ffn = nn.Sequential(
            nn.Linear(2*embed_dim, 2*embed_dim),
            nn.LayerNorm(2*embed_dim, elementwise_affine=True),
            nn.GELU(),
            nn.Linear(2*embed_dim, embed_dim)
        )

        if flash:
            self.flash = FastAttention(dim_head)
        else:
            self.flash = None

    def map_(self, func: Callable, x0: torch.Tensor, x1: torch.Tensor):
        return func(x0), func(x1)

    def forward(self, x0: torch.Tensor, x1: torch.Tensor) -> List[torch.Tensor]:
        qk0, qk1 = self.map_(self.to_qk, x0, x1)
        v0, v1 = self.map_(self.to_v, x0, x1)
        qk0, qk1, v0, v1 = map(
            lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.heads),
            (qk0, qk1, v0, v1))
        if self.flash is not None:
            m0 = self.flash(qk0, qk1, v1)
            m1 = self.flash(qk1, qk0, v0)
        else:
            qk0, qk1 = qk0 * self.scale**0.5, qk1 * self.scale**0.5
            sim = torch.einsum('b h i d, b h j d -> b h i j', qk0, qk1)
            attn01 = F.softmax(sim, dim=-1)
            attn10 = F.softmax(sim.transpose(-2, -1).contiguous(), dim=-1)
            m0 = torch.einsum('bhij, bhjd -> bhid', attn01, v1)
            m1 = torch.einsum('bhji, bhjd -> bhid', attn10.transpose(-2, -1), v0)
        m0, m1 = self.map_(lambda t: rearrange(t, 'b h n d -> b n (h d)'), m0, m1)
        m0, m1 = self.map_(self.to_out, m0, m1)
        x0 = x0 + self.ffn(torch.cat([x0, m0], -1))
        x1 = x1 + self.ffn(torch.cat([x1, m1], -1))
        return x0, x1


def sigmoid_log_double_softmax(
        sim: torch.Tensor, z0: torch.Tensor, z1: torch.Tensor) -> torch.Tensor:
    """ create the log assignment matrix from logits and similarity"""
    b, m, n = sim.shape
    certainties = F.logsigmoid(z0) + F.logsigmoid(z1).transpose(1, 2)
    scores0 = F.log_softmax(sim, 2)
    scores1 = F.log_softmax(
        sim.transpose(-1, -2).contiguous(), 2).transpose(-1, -2)
    scores = sim.new_full((b, m+1, n+1), 0)
    scores[:, :m, :n] = (scores0 + scores1 + certainties)
    scores[:, :-1, -1] = F.logsigmoid(-z0.squeeze(-1))
    scores[:, -1, :-1] = F.logsigmoid(-z1.squeeze(-1))
    return scores


class MatchAssignment(nn.Module):
    def __init__(self, dim: float) -> None:
        super(MatchAssignment, self).__init__()
        self.dim = dim
        self.matchability = nn.Linear(dim, 1, bias=True)
        self.final_proj = nn.Linear(dim, dim, bias=True)

    def forward(self, desc0: torch.Tensor, desc1: torch.Tensor):
        """ build assignment matrix from descriptors """
        mdesc0, mdesc1 = self.final_proj(desc0), self.final_proj(desc1)
        _, _, d = mdesc0.shape
        mdesc0, mdesc1 = mdesc0 / d**.25, mdesc1 / d**.25
        sim = torch.einsum('bmd,bnd->bmn', mdesc0, mdesc1)
        z0 = self.matchability(desc0)
        z1 = self.matchability(desc1)
        scores = sigmoid_log_double_softmax(sim, z0, z1)
        return scores, sim

    def scores(self, desc0: torch.Tensor, desc1: torch.Tensor):
        m0 = torch.sigmoid(self.matchability(desc0)).squeeze(-1)
        m1 = torch.sigmoid(self.matchability(desc1)).squeeze(-1)
        return m0, m1


def filter_matches(scores: torch.Tensor, th: float):
    """ obtain matches from a log assignment matrix [Bx M+1 x N+1]"""
    max0, max1 = scores[:, :-1, :-1].max(2), scores[:, :-1, :-1].max(1)
    m0, m1 = max0.indices, max1.indices
    mutual0 = torch.arange(m0.shape[1]).to(m0)[None] == m1.gather(1, m0)
    mutual1 = torch.arange(m1.shape[1]).to(m1)[None] == m0.gather(1, m1)
    max0_exp = max0.values.exp()
    zero = max0_exp.new_tensor(0)
    mscores0 = torch.where(mutual0, max0_exp, zero)
    mscores1 = torch.where(mutual1, mscores0.gather(1, m1), zero)
    if th is not None:
        valid0 = mutual0 & (mscores0 > th)
    else:
        valid0 = mutual0
    valid1 = mutual1 & valid0.gather(1, m1)
    m0 = torch.where(valid0, m0, m0.new_tensor(-1))
    m1 = torch.where(valid1, m1, m1.new_tensor(-1))
    return m0, m1, mscores0, mscores1


class LightGlue(nn.Module):
    default_conf = {
        'name': 'lightglue',  # just for interfacing
        'input_dim': 256,  # input descriptor dimension (autoselected from weights)
        'descriptor_dim': 256,
        'n_layers': 9,
        'num_heads': 4,
        'flash': False,  # enable FlashAttention
        'mp': False,  # enable mixed precision
        'filter_threshold': 0.1,  # match threshold
        'depth_confidence': -1,  # -1 is no early stopping, recommend: 0.95
        'width_confidence': -1,  # -1 is no point pruning, recommend: 0.99
        'weights': None,
    }

    required_data_keys = [
        'keypoints0', 'keypoints1', 'descriptors0', 'descriptors1']
    
    version = "v0.1_arxiv"
    url = "https://github.com/cvg/LightGlue/releases/download/{}/{}_lightglue.pth"

    pretrained = {
        'superpoint': ('superpoint_lightglue', 256),
        'disk': ('disk_lightglue', 128)
    }

    def __init__(self, pretrained='superpoint', **conf) -> None:
        super().__init__()
        self.conf = {**self.default_conf, **conf}
        if pretrained is not None:
            assert (pretrained in list(self.pretrained.keys()))
            self.conf['weights'], self.conf['input_dim'] = \
                self.pretrained[pretrained]
        self.conf = conf = SimpleNamespace(**self.conf)
            
        if conf.input_dim != conf.descriptor_dim:
            self.input_proj = nn.Linear(
                conf.input_dim, conf.descriptor_dim, bias=True)
        else:
            self.input_proj = nn.Identity()

        head_dim = conf.descriptor_dim // conf.num_heads
        self.posenc = LearnableFourierPositionalEncoding(2, head_dim, head_dim)

        h, n, d = conf.num_heads, conf.n_layers, conf.descriptor_dim
        self.self_attn = nn.ModuleList(
            [Transformer(d, h, conf.flash) for _ in range(n)])
        self.cross_attn = nn.ModuleList(
            [CrossTransformer(d, h, conf.flash) for _ in range(n)])
        self.log_assignment = nn.ModuleList(
            [MatchAssignment(d) for _ in range(n)])
        self.token_confidence = nn.ModuleList([
            TokenConfidence(d) for _ in range(n-1)])

        if pretrained is not None:
            fname = f'{conf.weights}_{self.version}.pth'.replace('.', '-')
            state_dict = torch.hub.load_state_dict_from_url(
                self.url.format(self.version, pretrained), file_name=fname)
            self.load_state_dict(state_dict, strict=False)
        elif conf.weights is not None:
            path = Path(__file__).parent
            path = path / 'weights/{}.pth'.format(self.conf.weights)
            state_dict = torch.load(str(path), map_location='cpu')
            self.load_state_dict(state_dict, strict=False)

        print('Loaded LightGlue model')

    def forward(self, data: dict) -> dict:
        """
        Match keypoints and descriptors between two images

        Input (dict):
            keypoints0: [B x M x 2], descriptors0: [B x M x D]
            keypoints1: [B x N x 2], descriptors1: [B x N x D]

        Output (dict):
            matches0: [B x M], matching_scores0: [B x M]
            matches1: [B x N], matching_scores1: [B x N]
            log_assignment: [B x M+1 x N+1]
        """
        with torch.autocast(enabled=self.conf.mp, device_type='cuda'):
            return self._forward(data)

    def _forward(self, data: dict) -> dict:
        for key in self.required_data_keys:
            assert key in data, f'Missing key {key} in data'
        kpts0_, kpts1_ = data['keypoints0'], data['keypoints1']
        b, m, _ = kpts0_.shape
        b, n, _ = kpts1_.shape

        kpts0 = normalize_keypoints(
            kpts0_, size=data.get('image_size0'), shape=None)
        kpts1 = normalize_keypoints(
            kpts1_, size=data.get('image_size1'), shape=None)

        assert torch.all(kpts0 >= -1) and torch.all(kpts0 <= 1)
        assert torch.all(kpts1 >= -1) and torch.all(kpts1 <= 1)

        desc0 = data['descriptors0'].detach().permute(0,2,1)
        desc1 = data['descriptors1'].detach().permute(0,2,1)

        assert(desc0.shape[-1] == self.conf.input_dim)
        assert(desc1.shape[-1] == self.conf.input_dim)

        if torch.is_autocast_enabled():
            desc0 = desc0.half()
            desc1 = desc1.half()

        desc0 = self.input_proj(desc0)
        desc1 = self.input_proj(desc1)

        # cache positional embeddings
        encoding0 = self.posenc(kpts0)
        encoding1 = self.posenc(kpts1)

        # GNN + final_proj + assignment
        ind0 = torch.arange(0, m).to(device=kpts0.device).expand(b,-1)
        ind1 = torch.arange(0, n).to(device=kpts0.device).expand(b,-1)
        prune0 = torch.ones_like(ind0)  # store layer where pruning is detected
        prune1 = torch.ones_like(ind1)
        dec, wic = self.conf.depth_confidence, self.conf.width_confidence
        token0, token1 = None, None
        for i in range(self.conf.n_layers):
            # self+cross attention
            desc0, desc1 = self.self_attn[i](
                desc0, desc1, encoding0, encoding1)
            desc0, desc1 = self.cross_attn[i](desc0, desc1)
            if i == self.conf.n_layers - 1:
                continue  # no early stopping or adaptive width at last layer
            if dec > 0:  # early stopping
                token0, token1 = self.token_confidence[i](desc0, desc1)
                if self.stop(token0, token1, self.conf_th(i), dec, m+n):
                    break
            if wic > 0:  # point pruning
                match0, match1 = self.log_assignment[i].scores(desc0, desc1)
                mask0 = self.get_mask(token0, match0, self.conf_th(i), 1-wic)
                mask1 = self.get_mask(token1, match1, self.conf_th(i), 1-wic)
                ind0, ind1 = ind0[mask0][None], ind1[mask1][None]
                desc0, desc1 = desc0[mask0][None], desc1[mask1][None]
                if desc0.shape[-2] == 0 or desc1.shape[-2] == 0:
                    break
                encoding0 = encoding0[:, :, mask0][:, None]
                encoding1 = encoding1[:, :, mask1][:, None]
            prune0[:, ind0] += 1
            prune1[:, ind1] += 1

        if wic > 0:  # scatter with indices after pruning
            scores_, _ = self.log_assignment[i](desc0, desc1)
            dt, dev = scores_.dtype, scores_.device
            scores = torch.zeros(b, m+1, n+1, dtype=dt, device=dev)
            scores[:, :-1, :-1] = -torch.inf
            scores[:, ind0[0], -1] = scores_[:, :-1, -1]
            scores[:, -1, ind1[0]] = scores_[:, -1, :-1]
            x, y = torch.meshgrid(ind0[0], ind1[0], indexing='ij')
            scores[:, x, y] = scores_[:, :-1, :-1]
        else:
            scores, _ = self.log_assignment[i](desc0, desc1)

        m0, m1, mscores0, mscores1 = filter_matches(
            scores, self.conf.filter_threshold)

        return {
            'log_assignment': scores,
            'matches0': m0,
            'matches1': m1,
            'matching_scores0': mscores0,
            'matching_scores1': mscores1,
            'stop': i+1,
            'prune0': prune0,
            'prune1': prune1,
        }

    def conf_th(self, i: int) -> float:
        """ scaled confidence threshold """
        return np.clip(
            0.8 + 0.1 * np.exp(-4.0 * i / self.conf.n_layers), 0, 1)

    def get_mask(self, confidence: torch.Tensor, match: torch.Tensor, conf_th: float, match_th: float) -> torch.Tensor:
        """ mask points which should be removed """
        if conf_th and confidence is not None:
            mask = torch.where(confidence > conf_th, match, match.new_tensor(1.0)) > match_th
        else:
            mask = match > match_th
        return mask

    def stop(self, token0: torch.Tensor, token1: torch.Tensor, conf_th: float, inl_th: float, seql: int) -> torch.Tensor:
        """ evaluate stopping condition"""
        tokens = torch.cat([token0, token1], -1)
        if conf_th:
            pos = 1.0 - (tokens < conf_th).float().sum() / seql
            return pos > inl_th
        else:
            return tokens.mean() > inl_th


from torch._six import string_classes
import collections.abc as collections
import torch


def map_tensor(input_, func):
    if isinstance(input_, torch.Tensor):
        return func(input_)
    elif isinstance(input_, string_classes):
        return input_
    elif isinstance(input_, collections.Mapping):
        return {k: map_tensor(sample, func) for k, sample in input_.items()}
    elif isinstance(input_, collections.Sequence):
        return [map_tensor(sample, func) for sample in input_]
    else:
        raise TypeError(
            f'input must be tensor, dict or list; found {type(input_)}')



import torch
import torch.nn as nn
import kornia
from types import SimpleNamespace


class DISK(nn.Module):
    default_conf = {
        'weights': 'depth',
        'max_num_keypoints': None,
        'desc_dim': 128,
        'nms_window_size': 5,
        'detection_threshold': 0.0,
        'pad_if_not_divisible': True,
    }
    required_data_keys = ['image']

    def __init__(self, **conf) -> None:
        super().__init__()
        self.conf = {**self.default_conf, **conf}
        self.conf = SimpleNamespace(**self.conf)
        self.model = kornia.feature.DISK.from_pretrained(self.conf.weights)

    def forward(self, data: dict) -> dict:
        image = data['image']

        features = self.model(
            image,
            n=self.conf.max_num_keypoints,
            window_size=self.conf.nms_window_size,
            score_threshold=self.conf.detection_threshold,
            pad_if_not_divisible=self.conf.pad_if_not_divisible
        )
        keypoints = [f.keypoints for f in features]
        scores = [f.detection_scores for f in features]
        descriptors = [f.descriptors for f in features]
        del features

        keypoints = torch.stack(keypoints, 0)
        scores = torch.stack(scores, 0)
        descriptors = torch.stack(descriptors, 0)

        return {
            'keypoints': keypoints.to(image),
            'keypoint_scores': scores.to(image),
            'descriptors': descriptors.to(image),
        }



from abc import ABCMeta, abstractmethod
from torch import nn
from copy import copy
import inspect


class BaseModel(nn.Module, metaclass=ABCMeta):
    default_conf = {}
    required_data_keys = []

    def __init__(self, conf):
        """Perform some logic and call the _init method of the child model."""
        super().__init__()
        self.conf = conf = {**self.default_conf, **conf}
        self.required_data_keys = copy(self.required_data_keys)
        self._init(conf)

    def forward(self, data):
        """Check the data and call the _forward method of the child model."""
        for key in self.required_data_keys:
            assert key in data, 'Missing key {} in data'.format(key)
        return self._forward(data)

    @abstractmethod
    def _init(self, conf):
        """To be implemented by the child class."""
        raise NotImplementedError

    @abstractmethod
    def _forward(self, data):
        """To be implemented by the child class."""
        raise NotImplementedError


def dynamic_load(root, model):
    module_path = f'{root.__name__}.{model}'
    module = __import__(module_path, fromlist=[''])
    classes = inspect.getmembers(module, inspect.isclass)
    # Filter classes defined in the module
    classes = [c for c in classes if c[1].__module__ == module_path]
    # Filter classes inherited from BaseModel
    classes = [c for c in classes if issubclass(c[1], BaseModel)]
    assert len(classes) == 1, classes
    return classes[0][1]
    # return getattr(module, 'Model')



import torch
import torch.nn.functional as F
import torch.nn as nn

class GeM(nn.Module):
    """Implementation of GeM as in https://github.com/filipradenovic/cnnimageretrieval-pytorch
    """
    def __init__(self, p=3, eps=1e-6):
        super().__init__()
        self.p = nn.Parameter(torch.ones(1)*p)
        self.eps = eps

    def forward(self, x):
        return F.avg_pool2d(x.clamp(min=self.eps).pow(self.p), (x.size(-2), x.size(-1))).pow(1./self.p)

class CosPlace(nn.Module):
    """
    CosPlace aggregation layer as implemented in https://github.com/gmberton/CosPlace/blob/main/model/network.py

    Args:
        in_dim: number of channels of the input
        out_dim: dimension of the output descriptor 
    """
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.gem = GeM()
        self.fc = nn.Linear(in_dim, out_dim)

    def forward(self, x):
        x = F.normalize(x, p=2, dim=1)
        x = self.gem(x)
        x = x.flatten(1)
        x = self.fc(x)
        x = F.normalize(x, p=2, dim=1)
        return x

if __name__ == '__main__':
    x = torch.randn(4, 2048, 10, 10)
    m = CosPlace(2048, 512)
    r = m(x)
    print(r.shape)


import torch
from torch import Tensor
from torch.nn import Module
from typing import Callable, Tuple
from .solver import find_fundamental_equirectangular,find_fundamental
from .metric import sampson
from typing import Optional,Union

def _batch_randperm(b: int, size: int, device: Optional[torch.device]= "cpu") -> Tensor:
    r"""Generate a batch of random permutations.

    Args:
        b (int): Batch size.
        size (int): Size of the permutations.
        device (str | torch.device, optional): Device to use. Defaults to 'cpu'.

    Returns:
        Tensor: A batch of random permutations.

    """

    perms = torch.argsort(torch.rand((b, size), device=device), dim=-1)
    return perms


def ransac(
    x: Tensor,
    y: Tensor,
    mask: Tensor,
    solver: Union[Callable,Module] = find_fundamental,
    evaluator: Union[Callable,Module] = sampson,
    ransac_ratio: float = 0.6,
    ransac_it: int = 8,
    ransac_thr: float = 0.8,
) -> Tuple[Tensor, Tensor, Tensor]:
    r"""RANSAC algorithm to find the best model.

    Args:
        x (Tensor): The first set of features with shape :math:`(B, N, D)`.
        y (Tensor): The second set of features with shape :math:`(B, M, D)`.
        mask (Tensor): The mask with shape :math:`(B, N, M)`.
        solver (Callable): The solver function to find the model.
        evaluator (Callable): The evaluator function to evaluate the model.
        ransac_ratio (float, optional): The ratio of inliers to consider the model as the best. Defaults to 0.6.
        ransac_it (int, optional): The number of iterations. Defaults to 16.
        ransac_thr (float, optional): The threshold to consider a point as an inlier. Defaults to 0.75.

    Returns:
        Tuple[Tensor, Tensor, Tensor]: The matching mask with shape :math:`(B, N, M)`.
    """

    B, N, D = x.shape
    B, M, D = y.shape

    r_n = int(ransac_ratio * N)
    r_m = int(ransac_ratio * M)
    perm1 = _batch_randperm(ransac_it, r_n, device=x.device).view(-1)
    perm2 = _batch_randperm(ransac_it, r_m, device=x.device).view(-1)

    s_x = x[:, perm1].view(B * ransac_it, r_n, D)
    s_y = y[:, perm2].view(B * ransac_it, r_m, D)
    s_m = mask[:, perm1].view(B, ransac_it, r_n, M)
    s_m = s_m.gather(-1, perm2.view(1, ransac_it, 1, r_m).repeat(B, 1, r_n, 1)).view(
        B * ransac_it, r_n, r_m
    )  # (B * ransac_it, N, M)

    models = solver(s_x, s_y, s_m)  # (B * ransac_it, D, D)
    x = x.unsqueeze(1).repeat(1, ransac_it, 1, 1).flatten(0, 1)
    y = y.unsqueeze(1).repeat(1, ransac_it, 1, 1).flatten(0, 1)
    mask = mask.unsqueeze(1).repeat(1, ransac_it, 1, 1).flatten(0, 1)
    errors = evaluator(models, x, y, mask)  # (B * ransac_it, N, M)
    errors = errors.view(B, ransac_it, N, M)
    models = models.view(B, ransac_it, models.shape[-2], models.shape[-1])
    avg_errors = torch.nanmean(errors, dim=(-1, -2))
    best_model_idx = torch.argmin(avg_errors, dim=-1)

    best_model = torch.gather(
        models,
        dim=1,
        index=best_model_idx.view(-1, 1, 1, 1).repeat(
            1, 1, models.shape[-2], models.shape[-1]
        ),
    ).squeeze(1)

    best_errors = torch.gather(
        errors,
        dim=1,
        index=best_model_idx.view(-1, 1, 1, 1).repeat(
            1, 1, errors.shape[-2], errors.shape[-1]
        ),
    ).squeeze(1)
    thrs = torch.nanquantile(
        best_errors.flatten(-2), ransac_thr, dim=-1, keepdim=True
    ).unsqueeze(-1)
    inliers = best_errors <= thrs
    best_errors[~inliers] = torch.nan
    return best_model, inliers, best_errors



import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import Callback, ModelCheckpoint
from torch.optim import lr_scheduler, optimizer
from unav.third_party.global_feature.mixVPR_main import utils

from unav.third_party.global_feature.mixVPR_main.models import helper


class VPRModel(pl.LightningModule):
    """This is the main model for Visual Place Recognition
    we use Pytorch Lightning for modularity purposes.

    Args:
        pl (_type_): _description_
    """

    def __init__(self,
                #---- Backbone
                backbone_arch='resnet50',
                pretrained=True,
                layers_to_freeze=1,
                layers_to_crop=[],
                
                #---- Aggregator
                agg_arch='ConvAP', #CosPlace, NetVLAD, GeM
                agg_config={},
                
                #---- Train hyperparameters
                lr=0.03, 
                optimizer='sgd',
                weight_decay=1e-3,
                momentum=0.9,
                warmpup_steps=500,
                milestones=[5, 10, 15],
                lr_mult=0.3,
                
                #----- Loss
                loss_name='MultiSimilarityLoss', 
                miner_name='MultiSimilarityMiner', 
                miner_margin=0.1,
                faiss_gpu=False
                 ):
        super().__init__()
        self.encoder_arch = backbone_arch
        self.pretrained = pretrained
        self.layers_to_freeze = layers_to_freeze
        self.layers_to_crop = layers_to_crop

        self.agg_arch = agg_arch
        self.agg_config = agg_config

        self.lr = lr
        self.optimizer = optimizer
        self.weight_decay = weight_decay
        self.momentum = momentum
        self.warmpup_steps = warmpup_steps
        self.milestones = milestones
        self.lr_mult = lr_mult

        self.loss_name = loss_name
        self.miner_name = miner_name
        self.miner_margin = miner_margin
        
        self.save_hyperparameters() # write hyperparams into a file
        
        self.loss_fn = utils.get_loss(loss_name)
        self.miner = utils.get_miner(miner_name, miner_margin)
        self.batch_acc = [] # we will keep track of the % of trivial pairs/triplets at the loss level 

        self.faiss_gpu = faiss_gpu
        
        # ----------------------------------
        # get the backbone and the aggregator
        self.backbone = helper.get_backbone(backbone_arch, pretrained, layers_to_freeze, layers_to_crop)
        self.aggregator = helper.get_aggregator(agg_arch, agg_config)
        
    # the forward pass of the lightning model
    def forward(self, x):
        x = torch.from_numpy(x).unsqueeze(0).permute(0, 3, 1, 2).float()
        print(x.size())
        x = self.backbone(x)
        print(x.size())
        x = self.aggregator(x)
        return x
    
    # configure the optimizer 
    def configure_optimizers(self):
        if self.optimizer.lower() == 'sgd':
            optimizer = torch.optim.SGD(self.parameters(), 
                                        lr=self.lr, 
                                        weight_decay=self.weight_decay, 
                                        momentum=self.momentum)
        elif self.optimizer.lower() == 'adamw':
            optimizer = torch.optim.AdamW(self.parameters(), 
                                        lr=self.lr, 
                                        weight_decay=self.weight_decay)
        elif self.optimizer.lower() == 'adam':
            optimizer = torch.optim.AdamW(self.parameters(), 
                                        lr=self.lr, 
                                        weight_decay=self.weight_decay)
        else:
            raise ValueError(f'Optimizer {self.optimizer} has not been added to "configure_optimizers()"')
        scheduler = lr_scheduler.MultiStepLR(optimizer, milestones=self.milestones, gamma=self.lr_mult)
        return [optimizer], [scheduler]
    
    # configure the optizer step, takes into account the warmup stage
    def optimizer_step(self,  epoch, batch_idx,
                        optimizer, optimizer_idx, optimizer_closure,
                        on_tpu, using_native_amp, using_lbfgs):
        # warm up lr
        if self.trainer.global_step < self.warmpup_steps:
            lr_scale = min(1., float(self.trainer.global_step + 1) / self.warmpup_steps)
            for pg in optimizer.param_groups:
                pg['lr'] = lr_scale * self.lr
        optimizer.step(closure=optimizer_closure)
        
    #  The loss function call (this method will be called at each training iteration)
    def loss_function(self, descriptors, labels):
        # we mine the pairs/triplets if there is an online mining strategy
        if self.miner is not None:
            miner_outputs = self.miner(descriptors, labels)
            loss = self.loss_fn(descriptors, labels, miner_outputs)
            
            # calculate the % of trivial pairs/triplets 
            # which do not contribute in the loss value
            nb_samples = descriptors.shape[0]
            nb_mined = len(set(miner_outputs[0].detach().cpu().numpy()))
            batch_acc = 1.0 - (nb_mined/nb_samples)

        else: # no online mining
            loss = self.loss_fn(descriptors, labels)
            batch_acc = 0.0
            if type(loss) == tuple: 
                # somes losses do the online mining inside (they don't need a miner objet), 
                # so they return the loss and the batch accuracy
                # for example, if you are developping a new loss function, you might be better
                # doing the online mining strategy inside the forward function of the loss class, 
                # and return a tuple containing the loss value and the batch_accuracy (the % of valid pairs or triplets)
                loss, batch_acc = loss

        # keep accuracy of every batch and later reset it at epoch start
        self.batch_acc.append(batch_acc)
        # log it
        self.log('b_acc', sum(self.batch_acc) /
                len(self.batch_acc), prog_bar=True, logger=True)
        return loss
    
    # This is the training step that's executed at each iteration
    def training_step(self, batch, batch_idx):
        places, labels = batch
        
        # Note that GSVCities yields places (each containing N images)
        # which means the dataloader will return a batch containing BS places
        BS, N, ch, h, w = places.shape
        
        # reshape places and labels
        images = places.view(BS*N, ch, h, w)
        labels = labels.view(-1)

        # Feed forward the batch to the model
        descriptors = self(images) # Here we are calling the method forward that we defined above
        loss = self.loss_function(descriptors, labels) # Call the loss_function we defined above
        
        self.log('loss', loss.item(), logger=True)
        return {'loss': loss}
    
    # This is called at the end of eatch training epoch
    def training_epoch_end(self, training_step_outputs):
        # we empty the batch_acc list for next epoch
        self.batch_acc = []

    # For validation, we will also iterate step by step over the validation set
    # this is the way Pytorch Lghtning is made. All about modularity, folks.
    def validation_step(self, batch, batch_idx, dataloader_idx=None):
        places, _ = batch
        # calculate descriptors
        descriptors = self(places)
        return descriptors.detach().cpu()
    
    def validation_epoch_end(self, val_step_outputs):
        """this return descriptors in their order
        depending on how the validation dataset is implemented 
        for this project (MSLS val, Pittburg val), it is always references then queries
        [R1, R2, ..., Rn, Q1, Q2, ...]
        """
        dm = self.trainer.datamodule
        # The following line is a hack: if we have only one validation set, then
        # we need to put the outputs in a list (Pytorch Lightning does not do it presently)
        if len(dm.val_datasets)==1: # we need to put the outputs in a list
            val_step_outputs = [val_step_outputs]
        
        for i, (val_set_name, val_dataset) in enumerate(zip(dm.val_set_names, dm.val_datasets)):
            feats = torch.concat(val_step_outputs[i], dim=0)
            
            if 'pitts' in val_set_name:
                # split to ref and queries
                num_references = val_dataset.dbStruct.numDb
                num_queries = len(val_dataset)-num_references
                positives = val_dataset.getPositives()
            elif 'msls' in val_set_name:
                # split to ref and queries
                num_references = val_dataset.num_references
                num_queries = len(val_dataset)-num_references
                positives = val_dataset.pIdx
            else:
                print(f'Please implement validation_epoch_end for {val_set_name}')
                raise NotImplemented

            r_list = feats[ : num_references]
            q_list = feats[num_references : ]
            pitts_dict = utils.get_validation_recalls(r_list=r_list, 
                                                q_list=q_list,
                                                k_values=[1, 5, 10, 15, 20, 50, 100],
                                                gt=positives,
                                                print_results=True,
                                                dataset_name=val_set_name,
                                                faiss_gpu=self.faiss_gpu
                                                )
            del r_list, q_list, feats, num_references, positives

            self.log(f'{val_set_name}/R1', pitts_dict[1], prog_bar=False, logger=True)
            self.log(f'{val_set_name}/R5', pitts_dict[5], prog_bar=False, logger=True)
            self.log(f'{val_set_name}/R10', pitts_dict[10], prog_bar=False, logger=True)
        print('\n\n')
            
            
if __name__ == '__main__':
    pl.utilities.seed.seed_everything(seed=190223, workers=True)
        
    datamodule = GSVCitiesDataModule(
        batch_size=120,
        img_per_place=4,
        min_img_per_place=4,
        shuffle_all=False, # shuffle all images or keep shuffling in-city only
        random_sample_from_each_place=True,
        image_size=(320, 320),
        num_workers=28,
        show_data_stats=True,
        val_set_names=['pitts30k_val', 'pitts30k_test', 'msls_val'], # pitts30k_val, pitts30k_test, msls_val
    )
    
    # examples of backbones
    # resnet18, resnet50, resnet101, resnet152,
    # resnext50_32x4d, resnext50_32x4d_swsl , resnext101_32x4d_swsl, resnext101_32x8d_swsl
    # efficientnet_b0, efficientnet_b1, efficientnet_b2
    # swinv2_base_window12to16_192to256_22kft1k
    model = VPRModel(
        #---- Encoder
        backbone_arch='resnet50',
        pretrained=True,
        layers_to_freeze=2,
        layers_to_crop=[4], # 4 crops the last resnet layer, 3 crops the 3rd, ...etc
        
        #---- Aggregator
        # agg_arch='CosPlace',
        # agg_config={'in_dim': 2048,
        #             'out_dim': 2048},
        # agg_arch='GeM',
        # agg_config={'p': 3},
        
        # agg_arch='ConvAP',
        # agg_config={'in_channels': 2048,
        #             'out_channels': 2048},

        agg_arch='MixVPR',
        agg_config={'in_channels' : 1024,
                'in_h' : 20,
                'in_w' : 20,
                'out_channels' : 1024,
                'mix_depth' : 4,
                'mlp_ratio' : 1,
                'out_rows' : 4}, # the output dim will be (out_rows * out_channels)
        
        #---- Train hyperparameters
        lr=0.05, # 0.0002 for adam, 0.05 or sgd (needs to change according to batch size)
        optimizer='sgd', # sgd, adamw
        weight_decay=0.001, # 0.001 for sgd and 0 for adam,
        momentum=0.9,
        warmpup_steps=650,
        milestones=[5, 10, 15, 25, 45],
        lr_mult=0.3,

        #----- Loss functions
        # example: ContrastiveLoss, TripletMarginLoss, MultiSimilarityLoss,
        # FastAPLoss, CircleLoss, SupConLoss,
        loss_name='MultiSimilarityLoss',
        miner_name='MultiSimilarityMiner', # example: TripletMarginMiner, MultiSimilarityMiner, PairMarginMiner
        miner_margin=0.1,
        faiss_gpu=False
    )
    
    # model params saving using Pytorch Lightning
    # we save the best 3 models accoring to Recall@1 on pittsburg val
    checkpoint_cb = ModelCheckpoint(
        monitor='pitts30k_val/R1',
        filename=f'{model.encoder_arch}' +
        '_epoch({epoch:02d})_step({step:04d})_R1[{pitts30k_val/R1:.4f}]_R5[{pitts30k_val/R5:.4f}]',
        auto_insert_metric_name=False,
        save_weights_only=True,
        save_top_k=3,
        mode='max',)

    #------------------
    # we instanciate a trainer
    trainer = pl.Trainer(
        accelerator='gpu', devices=[0],
        default_root_dir=f'./LOGS/{model.encoder_arch}', # Tensorflow can be used to viz 

        num_sanity_val_steps=0, # runs a validation step before stating training
        precision=16, # we use half precision to reduce  memory usage
        max_epochs=80,
        check_val_every_n_epoch=1, # run validation every epoch
        callbacks=[checkpoint_cb],# we only run the checkpointing callback (you can add more)
        reload_dataloaders_every_n_epochs=1, # we reload the dataset to shuffle the order
        log_every_n_steps=20,
        # fast_dev_run=True # uncomment or dev mode (only runs a one iteration train and validation, no checkpointing).
    )

    # we call the trainer, we give it the model and the datamodule
    trainer.fit(model=model, datamodule=datamodule)



import sys
from pathlib import Path

from unav.third_party.local_feature.SuperPoint_SuperGlue.base_model import BaseModel

sys.path.append(str(Path(__file__).parent))
from unav.third_party.local_feature.SuperPoint_SuperGlue.SuperGluePretrainedNetwork.models.superpoint import SuperPoint as SP


class SuperPoint(BaseModel):
    default_conf = {
        'nms_radius': 4,
        'keypoint_threshold': 0.005,
        'max_keypoints': -1,
        'remove_borders': 4,
    }
    required_inputs = ['image']

    def _init(self, conf):
        self.net = SP(conf)

    def _forward(self, data):
        return self.net(data)



from __future__ import annotations

from typing import Tuple, List

import torch
from kornia.color import rgb_to_grayscale
from kornia.feature import (
    BlobDoG,
    LAFOrienter,
    ScaleSpaceDetector,
    SIFTDescriptor,
    extract_patches_from_pyramid,
    get_laf_center,
)
import torchvision.transforms.functional as TF  # type: ignore
from PIL import Image
from kornia.geometry import ConvQuadInterp3d, ScalePyramid
from torch.nn import Module
import torch.nn.functional as F


class SIFT(Module):
    def __init__(
        self,
        num_features: int = 512,
        patch_size: int = 41,
        angle_bins: int = 8,
        spatial_bins: int = 8,
        **kwargs,
    ) -> None:
        super().__init__()
        self.patch_size = patch_size
        self.descriptor = SIFTDescriptor(
            patch_size, angle_bins, spatial_bins, rootsift=True
        )
        self.detector = ScaleSpaceDetector(
            num_features,
            resp_module=BlobDoG(),
            scale_space_response=True,  # We need that, because DoG operates on scale-space
            nms_module=ConvQuadInterp3d(10),
            scale_pyr_module=ScalePyramid(3, 1.6, patch_size, double_image=True),
            ori_module=LAFOrienter(19),
            mr_size=6.0,
            minima_are_also_good=True,
        )

    def detect(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        self.detector.to(x.device)
        with torch.no_grad():
            lafs, resps = self.detector(x.contiguous())
            return lafs, resps

    def describe(self, x: torch.Tensor, lafs: torch.Tensor) -> torch.Tensor:
        self.descriptor.to(x.device)
        with torch.no_grad():
            patches = extract_patches_from_pyramid(x, lafs, self.patch_size)
            B, N, CH, H, W = patches.size()
            descs = self.descriptor(patches.view(B * N, CH, H, W)).view(B, N, -1)
            descs = F.normalize(descs, dim=-1, p=2)
            return descs  # type: ignore

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        B, C, H, W = x.shape
        x = rgb_to_grayscale(x).float()
        lafs, resps = self.detect(x)
        descs = self.describe(x, lafs)
        kpts = get_laf_center(lafs)
        x = kpts[..., 0]
        y = kpts[..., 1]
        x = x / W * 2 - 1
        y = y / H * 2 - 1
        kpts = torch.stack([x, y], dim=-1)
        return kpts, descs


def detect(
    paths: List[str], num_features: int = 512
) -> Tuple[torch.Tensor, torch.Tensor]:
    descs = []
    kpts = []
    sift = SIFT(num_features=num_features).cuda()
    for path in paths:
        img = Image.open(path)
        img = TF.pil_to_tensor(img).float().cuda().unsqueeze(0)
        k, d = sift(img)
        descs.append(d.cpu())
        kpts.append(k.cpu())
    descs_t = torch.cat(descs, dim=0)
    kpts_t = torch.cat(kpts, dim=0)
    return kpts_t, descs_t



import glob
import os
from typing import Tuple

import torch
from PIL import Image
from torch.utils import data
import numpy as np
import torchvision.transforms as tvf
from tqdm import tqdm
import cv2

from main import VPRModel


class BaseDataset(data.Dataset):
    """Dataset with images from database and queries, used for inference (testing and building cache).
    """

    def __init__(self, img_path):
        super().__init__()
        self.img_path = img_path

        # path to images
        if 'query' in self.img_path:
            img_path_list = glob.glob(self.img_path + '/**/**/*.jpg', recursive=True)
            self.img_path_list = img_path_list
        elif 'db' in self.img_path:
            img_path_list = glob.glob(self.img_path + '/**/**/*.jpg', recursive=True)
            # sort images for db
            self.img_path_list = sorted(img_path_list, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        else:
            raise ValueError('img_path should be either query or db')
        assert len(self.img_path_list) > 0, f'No images found in {self.img_path}'

    def __getitem__(self, index):
        img = load_image(self.img_path_list[index])
        return img, index

    def __len__(self):
        return len(self.img_path_list)


class InferencePipeline:
    def __init__(self, model, dataset, feature_dim, batch_size=4, num_workers=4, device='cuda'):
        self.model = model
        self.dataset = dataset
        self.feature_dim = feature_dim
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.device = device

        self.dataloader = data.DataLoader(self.dataset,
                                          batch_size=self.batch_size,
                                          shuffle=False,
                                          num_workers=self.num_workers,
                                          pin_memory=True,
                                          drop_last=False)

    def run(self, split: str = 'db') -> np.ndarray:

        if os.path.exists(f'./LOGS/global_descriptors_{split}.npy'):
            print(f"Skipping {split} features extraction, loading from cache")
            return np.load(f'./LOGS/global_descriptors_{split}.npy')

        self.model.to(self.device)
        with torch.no_grad():
            global_descriptors = np.zeros((len(self.dataset), self.feature_dim))
            for batch in tqdm(self.dataloader, ncols=100, desc=f'Extracting {split} features'):
                imgs, indices = batch
                imgs = imgs.to(self.device)

                # model inference
                descriptors = self.model(imgs)
                descriptors = descriptors.detach().cpu().numpy()

                # add to global descriptors
                global_descriptors[np.array(indices), :] = descriptors

        # save global descriptors
        np.save(f'./LOGS/global_descriptors_{split}.npy', global_descriptors)
        return global_descriptors


def load_image(path):
    image_pil = Image.open(path).convert("RGB")

    # add transforms
    transforms = tvf.Compose([
        tvf.Resize((320, 320), interpolation=tvf.InterpolationMode.BICUBIC),
        tvf.ToTensor(),
        tvf.Normalize([0.485, 0.456, 0.406],
                      [0.229, 0.224, 0.225])
    ])

    # apply transforms
    image_tensor = transforms(image_pil)
    return image_tensor


def load_model(ckpt_path):
    # Note that images must be resized to 320x320
    model = VPRModel(backbone_arch='resnet50',
                     layers_to_crop=[4],
                     agg_arch='MixVPR',
                     agg_config={'in_channels': 1024,
                                 'in_h': 20,
                                 'in_w': 20,
                                 'out_channels': 1024,
                                 'mix_depth': 4,
                                 'mlp_ratio': 1,
                                 'out_rows': 4},
                     )

    state_dict = torch.load(ckpt_path)
    model.load_state_dict(state_dict)

    model.eval()
    print(f"Loaded model from {ckpt_path} Successfully!")
    return model


def calculate_top_k(q_matrix: np.ndarray,
                    db_matrix: np.ndarray,
                    top_k: int = 10) -> np.ndarray:
    # compute similarity matrix
    similarity_matrix = np.matmul(q_matrix, db_matrix.T)  # shape: (num_query, num_db)

    # compute top-k matches
    top_k_matches = np.argsort(-similarity_matrix, axis=1)[:, :top_k]  # shape: (num_query_images, 10)

    return top_k_matches


def record_matches(top_k_matches: np.ndarray,
                   query_dataset: BaseDataset,
                   database_dataset: BaseDataset,
                   out_file: str = 'record.txt') -> None:
    with open(f'{out_file}', 'a') as f:
        for query_index, db_indices in enumerate(tqdm(top_k_matches, ncols=100, desc='Recording matches')):
            pred_query_path = query_dataset.img_path_list[query_index]
            for i in db_indices.tolist():
                pred_db_paths = database_dataset.img_path_list[i]
            f.write(f'{pred_query_path} {pred_db_paths}\n')


def visualize(top_k_matches: np.ndarray,
              query_dataset: BaseDataset,
              database_dataset: BaseDataset,
              visual_dir: str = './LOGS/visualize',
              img_resize_size: Tuple = (320, 320)) -> None:
    if not os.path.exists(visual_dir):
        os.makedirs(visual_dir)
    for q_idx, db_idx in enumerate(tqdm(top_k_matches, ncols=100, desc='Visualizing matches')):
        pred_q_path = query_dataset.img_path_list[q_idx]
        q_array = cv2.imread(pred_q_path, cv2.IMREAD_COLOR)
        q_array = cv2.resize(q_array, img_resize_size, interpolation=cv2.INTER_CUBIC)
        gap_array = np.ones((q_array.shape[0], 10, 3)) * 255  # white gap

        for i in db_idx.tolist():
            pred_db_paths = database_dataset.img_path_list[i]
            db_array = cv2.imread(pred_db_paths, cv2.IMREAD_COLOR)
            db_array = cv2.resize(db_array, img_resize_size, interpolation=cv2.INTER_CUBIC)

            q_array = np.concatenate((q_array, gap_array, db_array), axis=1)

        result_array = q_array.astype(np.uint8)
        # result_array = cv2.cvtColor(result_array, cv2.COLOR_RGB2BGR)

        # save result as image using cv2
        cv2.imwrite(f'{visual_dir}/{os.path.basename(pred_q_path)}', result_array)


def main():
    # load images
    query_path = ''         # path to query images folder path
    datasets_path = ''      # path to database images folder path

    assert query_path == '' and datasets_path == '', 'Please specify the path to the query and datasets'

    query_dataset = BaseDataset(query_path)
    database_dataset = BaseDataset(datasets_path)

    # load model
    model = load_model('./LOGS/resnet50_MixVPR_4096_channels(1024)_rows(4).ckpt')

    # set up inference pipeline
    database_pipeline = InferencePipeline(model=model, dataset=database_dataset, feature_dim=4096)
    query_pipeline = InferencePipeline(model=model, dataset=query_dataset, feature_dim=4096)

    # run inference
    db_global_descriptors = database_pipeline.run(split='db')  # shape: (num_db, feature_dim)
    query_global_descriptors = query_pipeline.run(split='query')  # shape: (num_query, feature_dim)

    # calculate top-k matches
    top_k_matches = calculate_top_k(q_matrix=query_global_descriptors, db_matrix=db_global_descriptors, top_k=10)

    # record query_database_matches
    record_matches(top_k_matches, query_dataset, database_dataset, out_file='./LOGS/record.txt')

    # visualize top-k matches
    visualize(top_k_matches, query_dataset, database_dataset, visual_dir='./LOGS/visualize')


if __name__ == '__main__':
    main()



# import sys
from os.path import join,exists, isfile
# sys.path.append(join(sys.path[0],'third_party','pytorch_NetVlad'))
from unav.third_party.global_feature.pytorch_NetVlad.netvlad import NetVLAD
import torchvision.transforms as transforms
import json
import torch
import torchvision.models as models
import torch.nn as nn

def input_transform():
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406],
                             std=[0.229, 0.224, 0.225]),
    ])

class NetVladFeatureExtractor:
    def __init__(self, ckpt_path, arch='vgg16', num_clusters=64, pooling='netvlad', vladv2=False, nocuda=False,
                 input_transform=input_transform()):
        self.input_transform = input_transform

        flag_file = join(ckpt_path, 'checkpoints', 'flags.json')
        if exists(flag_file):
            with open(flag_file, 'r') as f:
                stored_flags = json.load(f)
                stored_num_clusters = stored_flags.get('num_clusters')
                if stored_num_clusters is not None:
                    num_clusters = stored_num_clusters
                    print(f'restore num_clusters to : {num_clusters}')
                stored_pooling = stored_flags.get('pooling')
                if stored_pooling is not None:
                    pooling = stored_pooling
                    print(f'restore pooling to : {pooling}')

        cuda = not nocuda
        if cuda and not torch.cuda.is_available():
            raise Exception("No GPU found, please run with --nocuda")

        self.device = torch.device("cuda" if cuda else "cpu")

        print('===> Building model')

        if arch.lower() == 'alexnet':
            encoder_dim = 256
            encoder = models.alexnet(pretrained=True)
            # capture only features and remove last relu and maxpool
            layers = list(encoder.features.children())[:-2]

            # if using pretrained only train conv5
            for l in layers[:-1]:
                for p in l.parameters():
                    p.requires_grad = False

        elif arch.lower() == 'vgg16':
            encoder_dim = 512
            encoder = models.vgg16(pretrained=True)
            # capture only feature part and remove last relu and maxpool
            layers = list(encoder.features.children())[:-2]

            # if using pretrained then only train conv5_1, conv5_2, and conv5_3
            for l in layers[:-5]:
                for p in l.parameters():
                    p.requires_grad = False

        encoder = nn.Sequential(*layers)
        self.model = nn.Module()
        self.model.add_module('encoder', encoder)

        if pooling.lower() == 'netvlad':
            net_vlad = NetVLAD(num_clusters=num_clusters, dim=encoder_dim, vladv2=vladv2)
            self.model.add_module('pool', net_vlad)
        else:
            raise ValueError('Unknown pooling type: ' + pooling)

        resume_ckpt = join(ckpt_path, 'checkpoints', 'checkpoint.pth.tar')
        print(resume_ckpt)

        if isfile(resume_ckpt):
            print("=> loading checkpoint '{}'".format(resume_ckpt))
            checkpoint = torch.load(resume_ckpt, map_location=lambda storage, loc: storage)
            best_metric = checkpoint['best_score']
            self.model.load_state_dict(checkpoint['state_dict'], strict=False)
            self.model = self.model.eval().to(self.device)
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(resume_ckpt, checkpoint['epoch']))
        else:
            print("=> no checkpoint found at '{}'".format(resume_ckpt))

    def feature(self, image):
        if self.input_transform:
            image = self.input_transform(image).unsqueeze(0)

        with torch.no_grad():
            input = image.to(self.device)
            image_encoding = self.model.encoder(input)
            vlad_encoding = self.model.pool(image_encoding)

            del input
            torch.cuda.empty_cache()
            return vlad_encoding


import torch
import torch.nn as nn
import timm
import numpy as np


class Swin(nn.Module):
    def __init__(self,
                 model_name='swinv2_base_window12to16_192to256_22kft1k',
                 pretrained=True,
                 layers_to_freeze=2
                 ):
        """Class representing the Swin (V1 and V2) backbone used in the pipeline
        Swin contains 4 layers (0 to 3), where layer 2 is the heaviest in terms of # params

        Args:
            model_name (str, optional): The architecture of the Swin backbone to instanciate. Defaults to 'swinv2_base_window12to16_192to256_22kft1k'.
            pretrained (bool, optional): Whether pretrained or not. Defaults to True.
            layers_to_freeze (int, optional): The number of blocks to freeze in layers[2]  (starting from 0) . Defaults to 2.
        """
        super().__init__()
        self.model_name = model_name
        self.layers_to_freeze = layers_to_freeze        
        self.model = timm.create_model(model_name, pretrained=pretrained, num_classes=0)
        self.model.head = None
        
        if pretrained:
            self.model.patch_embed.requires_grad_(False)
            self.model.layers[0].requires_grad_(False)
            self.model.layers[1].requires_grad_(False)
            # layers[2] contains most of the blocks, better freeze some of them
            for i in range(layers_to_freeze*5): # we make 5 steps (swin contains lots of layers)
                self.model.layers[2].blocks[i].requires_grad_(False)
            

        if 'base' in model_name:
            out_channels = 1024
        elif 'large' in model_name:
            out_channels = 1536
        else:
            out_channels = 768
        self.out_channels = out_channels
        
        if '384' in model_name:
            self.depth = 144
        else:
            self.depth = 49
            
    def forward(self, x):
        x = self.model.forward_features(x)
        # the following is a hack to make the output of the transformer
        # as a 3D feature maps
        bs, f, c = x.shape
        x = x.view(bs, int(np.sqrt(f)), int(np.sqrt(f)), c)
        return x.permute(0,3,1,2)
    

def print_nb_params(m):
    model_parameters = filter(lambda p: p.requires_grad, m.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    print(f'Trainable parameters: {params/1e6:.3}M')

if __name__ == '__main__':
    x = torch.randn(4,3,256,256)
    m = Swin(model_name='swinv2_base_window12to16_192to256_22kft1k',
                 pretrained=True,
                 layers_to_freeze=2,)
    r = m(x)
    print_nb_params(m)
    print(f'Input shape is {x.shape}')
    print(f'Output shape is {r.shape}')


import torch
import h5py
import numpy as np

hfile = h5py.File('/home/endeleze/Desktop/WeNev/Localization/pytorch-NetVlad/Out/GS010021/global_features.h5', 'r')
q_hfile = h5py.File('/home/endeleze/Desktop/WeNev/Localization/pytorch-NetVlad/Out/GS010021/query/global_features.h5', 'r')

names = []
hfile.visititems(
    lambda _, obj: names.append(obj.parent.name.strip('/'))
    if isinstance(obj, h5py.Dataset) else None)
names = list(set(names))
db_names = [n for n in names]
q_names = []
q_hfile.visititems(
    lambda _, obj: q_names.append(obj.parent.name.strip('/'))
    if isinstance(obj, h5py.Dataset) else None)
q_names = list(set(q_names))
query_names = [n for n in q_names]


def tensor_from_names(names, hfile):
    desc = [hfile[i]['global_descriptor'].__array__() for i in names]
    desc = torch.from_numpy(np.stack(desc, 0)).to('cuda').float()
    return desc


db_desc = tensor_from_names(db_names, hfile)
query_desc = tensor_from_names(query_names, q_hfile)
sim = torch.einsum('id,jd->ij', query_desc, db_desc)
topk = torch.topk(sim, 10, dim=1).indices.cpu().numpy()

pairs = []
for query, indices in zip(query_names, topk):
    for i in indices:
        if query != db_names[i] :
            pair = (query, db_names[i])
            pairs.append(pair)

with open('/home/endeleze/Desktop/WeNev/Localization/pytorch-NetVlad/Out/GS010021/pairs.txt', 'w') as f:
    f.write('\n'.join(' '.join([i, j]) for i, j in pairs))


from typing import List

import torch
from PIL import Image
from torchvision.transforms.functional import pil_to_tensor, to_pil_image
from torchvision.utils import draw_keypoints


def visualize_keypoints(image, keypoints, color="red", radius=2):
    image = draw_keypoints(image, keypoints, colors=color, radius=radius)
    image = to_pil_image(image)
    return image


def draw_match_lines(image1, image2, kps1, kps2):
    image1 = image1.cpu()
    image2 = image2.cpu()
    kps1 = kps1.cpu()
    kps2 = kps2.cpu()
    kps2 = kps2 + torch.tensor([image1.shape[-1], 0])
    all_kps = torch.cat([kps1, kps2], dim=0)
    all_kps = all_kps.unsqueeze(0)
    connectivity = torch.stack(
        [torch.arange(len(kps1)), torch.arange(len(kps2)) + len(kps1)], dim=1
    )
    connectivity = connectivity.tolist()

    side_by_side = torch.cat([image1, image2], dim=-1)
    image = draw_keypoints(side_by_side, all_kps, connectivity, colors="red", radius=2)
    image = to_pil_image(image)
    return image


def draw_transfrom_points(
    index_image,
    query_image,
    H,
    selected_X,
    selected_Y,
):

    padded_X = torch.cat(
        [selected_X, torch.ones(selected_X.shape[0], 1, device=selected_X.device)],
        dim=-1,
    )

    projected_X = torch.matmul(padded_X, H.transpose(-1, -2))
    projected_X = projected_X[..., :2] / projected_X[..., 2:]

    index_image = visualize_keypoints(index_image, selected_X.unsqueeze(0))
    query_image = visualize_keypoints(query_image, selected_Y.unsqueeze(0))
    query_image = pil_to_tensor(query_image)

    query_image = visualize_keypoints(query_image, projected_X, color="blue")
    side_by_side = concat(index_image, query_image, dim=1)
    return side_by_side


def concat(im1, im2, dim=1):
    dest = None
    if dim == 1:
        dest = Image.new("RGB", (im1.width + im2.width, im1.height))
        dest.paste(im1, (0, 0))
        dest.paste(im2, (im1.width, 0))
    else:
        dest = Image.new("RGB", (im1.width, im1.height + im2.height))
        dest.paste(im1, (0, 0))
        dest.paste(im2, (0, im1.height))
    return dest
